# Utilities related to the clipboard
#
# Copyright 2024-2025 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Utilities related to the clipboard."""

from __future__ import annotations

import re
import time
from typing import TYPE_CHECKING, Any

import gi
from dasbus.connection import SessionMessageBus
from dasbus.error import DBusError

gi.require_version("Atspi", "2.0")
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Atspi, Gdk, Gtk

from . import (
    clipboard_command_definitions,
    dbus_service,
    debug,
    guilabels,
    input_event,
    input_event_manager,
    messages,
    presentation_manager,
    script_manager,
)
from .ax_utilities import AXUtilities
from .extension import Extension

if TYPE_CHECKING:
    from collections.abc import Callable

    from dasbus.client.proxy import InterfaceProxy

    from .command import Command
    from .scripts import default


class _ClipboardManager:
    """Base class for interacting with clipboard managers."""

    def __init__(self, name: str, change_callback: Callable[[str], None]) -> None:
        self._name: str = name
        self._change_callback: Callable[[str], None] = change_callback
        self._contents: str = ""
        self._is_active: bool = False

    def is_active(self) -> bool:
        """Returns True if this manager is active."""

        return self._is_active

    def connect(self) -> None:
        """Connects to the clipboard manager."""

    def disconnect(self) -> None:
        """Disconnects from the clipboard manager."""

    def set_contents(self, text: str) -> None:
        """Sets the contents of the clipboard to text."""

    def get_contents(self) -> str:
        """Returns the pre-stored contents of the clipboard."""

        if not self._contents:
            self._contents = self._get_contents()
        return self._contents

    def _get_contents(self) -> str:
        """Obtains and returns the contents of the clipboard."""

        return ""

    def _on_contents_changed(self, *args: tuple[Any, ...], **kwargs: dict[str, Any]) -> None:
        """Notifies the registered callback that the contents changed."""

        msg = f"{self._name}: Contents changed. {args} {kwargs}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._contents = self._get_contents()
        self._change_callback(self._contents)


class _ClipboardManagerFallback(_ClipboardManager):
    """Class for interacting with the clipboard via Gtk.Clipboard."""

    def __init__(self, change_callback: Callable[[str], None]) -> None:
        super().__init__("FALLBACK", change_callback)
        self._handler_id: int | None = None

    def connect(self) -> None:
        """Connects to the clipboard manager."""

        if self._handler_id is not None:
            return

        clipboard = Gtk.Clipboard.get(Gdk.Atom.intern("CLIPBOARD", False))
        self._handler_id = clipboard.connect("owner-change", self._on_contents_changed)
        self._is_active = True

    def disconnect(self) -> None:
        """Disconnects from the clipboard manager."""

        self._is_active = False
        if self._handler_id is None:
            return

        clipboard = Gtk.Clipboard.get(Gdk.Atom.intern("CLIPBOARD", False))
        clipboard.disconnect(self._handler_id)
        self._handler_id = None

    def _get_contents(self) -> str:
        """Obtains and returns the contents of the clipboard."""

        if self._handler_id is None:
            return ""

        clipboard = Gtk.Clipboard.get(Gdk.Atom.intern("CLIPBOARD", False))
        result = clipboard.wait_for_text()
        if result is None:
            msg = "FALLBACK: Have handler, but text is None"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        debug_string = result.replace("\n", "\\n")
        msg = f"FALLBACK: Clipboard contents: {debug_string}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def set_contents(self, text: str) -> None:
        """Sets the contents of the clipboard to text."""

        msg = f"FALLBACK: Setting clipboard contents to: {text}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        clipboard = Gtk.Clipboard.get(Gdk.Atom.intern("CLIPBOARD", False))
        clipboard.set_text(text, -1)


class _ClipboardManagerGPaste(_ClipboardManager):
    """Class for interacting with the clipboard via GPaste."""

    def __init__(self, change_callback: Callable[[str], None]) -> None:
        super().__init__("GPASTE", change_callback)
        self._bus: SessionMessageBus | None = None
        self._gpaste_proxy: InterfaceProxy | None = None
        self._props_proxy: InterfaceProxy | None = None
        self._signal_subscription: Any | None = None
        self._original_active_state: bool | None = None

    def connect(self) -> None:
        """Connects to the clipboard manager."""

        try:
            self._bus = SessionMessageBus()
            self._gpaste_proxy = self._bus.get_proxy("org.gnome.GPaste", "/org/gnome/GPaste")

            # Test if the service is actually available by checking properties
            self._props_proxy = self._bus.get_proxy(
                "org.gnome.GPaste",
                "/org/gnome/GPaste",
                "org.freedesktop.DBus.Properties",
            )
            self._original_active_state = self._props_proxy.Get("org.gnome.GPaste2", "Active")
        except DBusError as error:
            msg = f"CLIPBOARD PRESENTER: Could not access GPaste interface: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        try:
            self._original_active_state = self._props_proxy.Get("org.gnome.GPaste2", "Active")
            if not self._original_active_state:
                msg = "CLIPBOARD PRESENTER: GPaste is not active. Enabling Tracking."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self._gpaste_proxy.Track(True)
                new_state = self._props_proxy.Get("org.gnome.GPaste2", "Active")
                msg = f"CLIPBOARD PRESENTER: Is active now: {bool(new_state)}"
                debug.print_message(debug.LEVEL_INFO, msg, True)

            self._signal_subscription = self._gpaste_proxy.Update.connect(self._on_contents_changed)
            self._is_active = True
        except DBusError as error:
            msg = f"CLIPBOARD PRESENTER: Could not connect to GPaste signals: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._gpaste_proxy = None
            self._props_proxy = None
            self._bus = None

    def disconnect(self) -> None:
        """Disconnects from the clipboard manager."""

        if self._gpaste_proxy is None or self._props_proxy is None:
            msg = "CLIPBOARD PRESENTER: Cannot disconnect due to missing interface(s)."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        if self._signal_subscription is not None:
            self._signal_subscription.disconnect()
            self._signal_subscription = None

        if not self._original_active_state:
            msg = "CLIPBOARD PRESENTER: Restoring inactive state by disabling tracking."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._gpaste_proxy.Track(False)
            new_state = self._props_proxy.Get("org.gnome.GPaste2", "Active")
            msg = f"CLIPBOARD PRESENTER: Is active now: {bool(new_state)}"
            debug.print_message(debug.LEVEL_INFO, msg, True)

        self._gpaste_proxy = None
        self._props_proxy = None
        self._bus = None
        self._is_active = False

    def _get_contents(self) -> str:
        """Obtains and returns the contents of the clipboard."""

        if self._gpaste_proxy is None:
            return ""

        try:
            result = self._gpaste_proxy.GetElementAtIndex(0)[1]
        except DBusError as error:
            msg = f"GPASTE: Could not get clipboard contents: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        debug_string = result.replace("\n", "\\n")
        msg = f"GPASTE: Clipboard contents: {debug_string}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def set_contents(self, text: str) -> None:
        """Sets the contents of the clipboard to text."""

        if self._gpaste_proxy is None:
            return

        self._gpaste_proxy.Add(text)


class _ClipboardManagerKlipper(_ClipboardManager):
    """Class for interacting with the clipboard via Klipper ."""

    def __init__(self, change_callback: Callable[[str], None]) -> None:
        super().__init__("KLIPPER", change_callback)
        self._bus: SessionMessageBus | None = None
        self._klipper_proxy: InterfaceProxy | None = None
        self._signal_subscription: Any | None = None

    def connect(self) -> None:
        """Connects to the clipboard manager."""

        try:
            self._bus = SessionMessageBus()
            self._klipper_proxy = self._bus.get_proxy("org.kde.klipper", "/klipper")

            # Test if the service is actually available by calling a simple method
            self._klipper_proxy.getClipboardContents()
        except DBusError as error:
            msg = f"CLIPBOARD PRESENTER: Could not access klipper interface: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        try:
            self._signal_subscription = self._klipper_proxy.clipboardHistoryUpdated.connect(
                self._on_contents_changed,
            )
            self._is_active = True
        except DBusError as error:
            msg = f"CLIPBOARD PRESENTER: Could not connect to klipper signal: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._klipper_proxy = None
            self._bus = None

    def disconnect(self) -> None:
        """Disconnects from the clipboard manager."""

        if self._signal_subscription is not None:
            self._signal_subscription.disconnect()
        self._signal_subscription = None
        self._klipper_proxy = None
        self._bus = None
        self._is_active = False

    def _get_contents(self) -> str:
        """Obtains and returns the contents of the clipboard."""

        if self._klipper_proxy is None:
            return ""

        result = self._klipper_proxy.getClipboardContents()
        debug_string = result.replace("\n", "\\n")
        msg = f"KLIPPER: Clipboard contents: {debug_string}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def set_contents(self, text: str) -> None:
        """Sets the contents of the clipboard to text."""

        if self._klipper_proxy is None:
            return

        msg = f"KLIPPER: Setting clipboard contents to: {text}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._klipper_proxy.setClipboardContents(text)


class ClipboardPresenter(Extension):
    """Manages clipboard-related functionality."""

    GROUP_LABEL = guilabels.KB_GROUP_CLIPBOARD

    def __init__(self) -> None:
        self._event_listener: Atspi.EventListener = Atspi.EventListener.new(self._listener)
        self._last_clipboard_update_text: str = ""
        self._last_clipboard_update_time: float = time.time()
        self._manager: _ClipboardManager | None = None
        super().__init__()

    def _get_commands(self) -> list[Command]:
        return clipboard_command_definitions.get_commands(self)

    @dbus_service.command
    def present_clipboard_contents(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the clipboard contents."""

        tokens = [
            "CLIPBOARD PRESENTER: present_clipboard_contents. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self._manager is None:
            msg = "CLIPBOARD PRESENTER: Cannot present contents, no active manager."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        contents = self._manager.get_contents()
        if not contents or len(contents) > 5000:
            contents = messages.character_count(len(contents))
        presentation_manager.get_manager().present_message(
            messages.CLIPBOARD_CONTAINS % contents,
            contents,
        )
        return True

    def _connect(self) -> None:
        """Connects to the clipboard manager."""

        if self._manager is not None:
            return

        manager: _ClipboardManager | None = None

        # If you try to connect to Klipper from a GNOME session, it will fail with a DBus
        # exception. However, if you try to connect to GPaste from a KDE session, it will
        # succeed -- or at least not throw an exception. Therefore, check for Klipper first.
        manager = _ClipboardManagerKlipper(self._present_clipboard_contents_change)
        manager.connect()
        if manager.is_active():
            self._manager = manager
            msg = "CLIPBOARD PRESENTER: Using Klipper."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        # See comment above. Check for GPaste last.
        manager = _ClipboardManagerGPaste(self._present_clipboard_contents_change)
        manager.connect()
        if manager.is_active():
            self._manager = manager
            msg = "CLIPBOARD PRESENTER: Using GPaste."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        self._manager = _ClipboardManagerFallback(self._present_clipboard_contents_change)
        self._manager.connect()
        msg = "CLIPBOARD PRESENTER: Using Gtk.Clipboard as fallback."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _disconnect(self) -> None:
        """Disconnects from the clipboard manager."""

        if self._manager is None:
            return
        self._manager.disconnect()

    def _get_contents(self) -> str:
        """Returns the clipboard contents."""

        if self._manager is None:
            msg = "CLIPBOARD PRESENTER: Cannot get contents, no active manager"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        result = self._manager.get_contents()
        debug_string = result.replace("\n", "\\n")
        msg = f"CLIPBOARD PRESENTER: Current contents: {debug_string}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def activate(self) -> None:
        """Activates the presenter."""

        debug.print_message(debug.LEVEL_INFO, "CLIPBOARD PRESENTER: Activating", True)
        self._event_listener.register("object:text-changed")
        self._connect()
        debug.print_message(debug.LEVEL_INFO, "CLIPBOARD PRESENTER: Activated", True)

    def deactivate(self) -> None:
        """Deactivates the presenter."""

        debug.print_message(debug.LEVEL_INFO, "CLIPBOARD PRESENTER: Deactivating", True)
        self._event_listener.deregister("object:text-changed")
        self._disconnect()
        debug.print_message(debug.LEVEL_INFO, "CLIPBOARD PRESENTER: Deactivated", True)

    def append_text(self, text: str, separator: str = "\n") -> None:
        """Appends text to the clipboard contents."""

        if self._manager is None:
            msg = "CLIPBOARD PRESENTER: Cannot append text, no active manager."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        old_text = self._manager.get_contents()
        new_text = f"{old_text}{separator}{text}"
        msg = f"CLIPBOARD PRESENTER: Appending '{text}'. New contents: '{new_text}'."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._manager.set_contents(new_text)

    def set_text(self, text: str) -> None:
        """Sets the clipboard contents to text."""

        if self._manager is None:
            msg = "CLIPBOARD PRESENTER: Cannot set text, no active manager."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        msg = f"CLIPBOARD PRESENTER: Setting text to '{text}'."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._manager.set_contents(text)

    def is_clipboard_text_changed_event(self, event: Atspi.Event) -> bool:
        """Returns True if event is a text changed event associated with the clipboard."""

        if not event.type.startswith("object:text-changed"):
            return False

        if not (AXUtilities.is_editable(event.source) or AXUtilities.is_terminal(event.source)):
            return False

        manager = input_event_manager.get_manager()
        if (
            not manager.last_event_was_command()
            or manager.last_event_was_undo()
            or manager.last_event_was_backspace()
        ):
            return False

        if "delete" in event.type and manager.last_event_was_paste():
            return False

        contents = self._get_contents()
        if not contents:
            return False

        # Some applications send multiple text insertion events for part of a given paste.
        words_match = bool(re.search(r"\w", event.any_data)) == bool(re.search(r"\w", contents))
        return words_match and contents.startswith(event.any_data.rstrip())

    def _present_clipboard_contents_change(self, string: str) -> None:
        """Presents the clipboard contents change."""

        msg = f"CLIPBOARD PRESENTER: Contents changed to: '{string}'"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if (
            string == self._last_clipboard_update_text
            and time.time() - self._last_clipboard_update_time < 1
        ):
            msg = "CLIPBOARD PRESENTER: Not presenting change: likely duplicate."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        self._last_clipboard_update_text = string
        self._last_clipboard_update_time = time.time()
        script = script_manager.get_manager().get_active_script()
        if script is None:
            return

        manager = input_event_manager.get_manager()
        if manager.last_event_was_cut():
            presentation_manager.get_manager().present_message(
                messages.CLIPBOARD_CUT_FULL,
                messages.CLIPBOARD_CUT_BRIEF,
            )
            return

        if manager.last_event_was_copy():
            presentation_manager.get_manager().present_message(
                messages.CLIPBOARD_COPIED_FULL,
                messages.CLIPBOARD_COPIED_BRIEF,
            )
            return

        if manager.last_event_was_paste():
            presentation_manager.get_manager().present_message(
                messages.CLIPBOARD_PASTED_FULL,
                messages.CLIPBOARD_PASTED_BRIEF,
            )
            return

        msg = "CLIPBOARD PRESENTER: Not presenting change: is not cut, copy, or paste"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _listener(self, event: Atspi.Event) -> None:
        """Generic listener for events of interest."""

        tokens = ["CLIPBOARD PRESENTER: Possible change event", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if self.is_clipboard_text_changed_event(event):
            self._present_clipboard_contents_change(event.any_data)


_presenter: ClipboardPresenter = ClipboardPresenter()


def get_presenter() -> ClipboardPresenter:
    """Returns the Clipboard Presenter singleton."""

    return _presenter



# Orca
#
# Copyright 2006-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for the pronunciation dictionary."""

from __future__ import annotations

from typing import TYPE_CHECKING

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gtk  # pylint: disable=no-name-in-module

from . import (
    gsettings_registry,
    guilabels,
    messages,
    orca_gui_helpers,
    preferences_grid_base,
    presentation_manager,
    script_manager,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from .pronunciation_dictionary_manager import PronunciationDictionaryManager
    from .scripts import default


class PronunciationDictionaryPreferencesGrid(  # pylint: disable=too-many-instance-attributes
    preferences_grid_base.PreferencesGridBase,
):
    """GtkGrid containing the Pronunciation Dictionary preferences page."""

    # pylint: disable=no-member

    BUILD_BATCH_SIZE: int = 50

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for pronunciation preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.PRONUNCIATION,
            panel_id="manual.pronunciation",
            description=(
                guilabels.PRONUNCIATION_DICTIONARY_INFO
                + "\n\n"
                + "The pronunciation dictionary is a list of words and replacement text. "
                "Add a new entry from the dictionary header. Edit or delete an existing "
                "entry from its row."
            ),
            schema="pronunciations",
            show_available_settings=False,
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.PRONUNCIATION_DICTIONARY,
                    kind="list",
                    summary=(
                        "Lists words and the replacement text Orca sends to speech before "
                        "speaking them."
                    ),
                    actions=(
                        preferences_grid_base.PreferenceActionDoc(
                            label=guilabels.ADD_NEW_PRONUNCIATION,
                            summary=(
                                "Opens a dialog where you can add a word and the replacement "
                                "text Orca should speak."
                            ),
                            placement="header",
                        ),
                    ),
                    item_actions=(
                        preferences_grid_base.PreferenceActionDoc(
                            label=guilabels.EDIT_PRONUNCIATION,
                            summary=(
                                "Opens a dialog where you can change the selected pronunciation "
                                "entry."
                            ),
                        ),
                        preferences_grid_base.PreferenceActionDoc(
                            label=guilabels.DICTIONARY_DELETE,
                            summary="Removes the selected pronunciation entry.",
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, manager: PronunciationDictionaryManager, script: default.Script) -> None:
        super().__init__(guilabels.PRONUNCIATION)
        self._manager: PronunciationDictionaryManager = manager
        self._script: default.Script = script
        self._initializing: bool = True

        self._pronunciations: list[tuple[str, str]] = []
        self._inherited_keys: set[str] = set()
        self._deleted_inherited_keys: set[str] = set()
        self._listbox: Gtk.ListBox | None = None
        self._has_unsaved_changes: bool = False
        self._loaded_from_settings: bool = False
        self._pending_build_id: int = 0
        self._pending_build_index: int = 0

        self._build()
        # Build rows lazily once the page is mapped in idle-time chunks. With thousands of entries,
        # mapping the page all at once floods AT-SPI with state-changed:showing events and blocks
        # main loop for several seconds. Chunking spreads that flood across idle ticks.
        self.connect("map", self._on_map)
        self.connect("destroy", self._on_destroy)
        self.refresh()

    def _build(self) -> None:
        """Create the Gtk widgets composing the grid."""

        row = 0

        # Info box
        info_listbox = orca_gui_helpers.create_info_listbox(guilabels.PRONUNCIATION_DICTIONARY_INFO)
        self.attach(info_listbox, 0, row, 1, 1)
        row += 1

        header_box, _add_button = orca_gui_helpers.create_heading_action_box(
            guilabels.PRONUNCIATION_DICTIONARY,
            "list-add-symbolic",
            guilabels.DICTIONARY_NEW_ENTRY,
            self._on_add_clicked,
        )

        self.attach(header_box, 0, row, 1, 1)
        row += 1

        # ListBox for pronunciation entries
        self._listbox = orca_gui_helpers.create_framed_listbox(
            selection_mode=Gtk.SelectionMode.SINGLE,
        )

        # Suppress pronunciation substitution while editing entries
        self._listbox.connect("realize", self._on_listbox_realize)
        self._listbox.connect("unrealize", self._on_listbox_unrealize)

        scrolled_window = orca_gui_helpers.create_preferences_scrolled_window(self._listbox)
        self.attach(scrolled_window, 0, row, 1, 1)

    def _create_two_label_row(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        left_label_text: str,
        right_label_text: str,
        edit_handler: Callable[[Gtk.Button], None] | None = None,
        delete_handler: Callable[[Gtk.Button], None] | None = None,
        include_top_separator: bool = True,
        left_label_size_group: Gtk.SizeGroup | None = None,
    ) -> Gtk.ListBoxRow:
        """Create a single listbox row with two labels and optional edit/delete buttons."""

        actions = []
        if delete_handler:
            actions.append(
                orca_gui_helpers.ListRowAction(
                    "delete",
                    guilabels.DICTIONARY_DELETE,
                    delete_handler,
                    "user-trash-symbolic",
                )
            )
        if edit_handler:
            actions.append(
                orca_gui_helpers.ListRowAction(
                    "edit",
                    guilabels.DIALOG_EDIT,
                    edit_handler,
                    "document-edit-symbolic",
                )
            )

        row, _left_label, _right_label, action_buttons = orca_gui_helpers.create_action_list_row(
            left_label_text,
            right_label_text,
            actions,
            include_top_separator=include_top_separator,
            primary_label_size_group=left_label_size_group,
        )

        row.delete_button = action_buttons.get("delete")
        row.edit_button = action_buttons.get("edit")

        return row

    def _create_row(
        self,
        phrase: str,
        substitution: str,
        row_index: int,
        include_top_separator: bool = True,
    ) -> Gtk.ListBoxRow:
        """Create a ListBoxRow for a pronunciation entry."""

        row = self._create_two_label_row(
            phrase,
            substitution,
            edit_handler=self._on_edit_clicked,
            delete_handler=self._on_delete_clicked,
            include_top_separator=include_top_separator,
        )

        row.pronunciation_row_index = row_index

        if row.edit_button:
            row.edit_button.pronunciation_row_index = row_index

        if row.delete_button:
            row.delete_button.pronunciation_row_index = row_index

        return row

    def _on_edit_clicked(self, button: Gtk.Button) -> None:
        """Handle edit button click."""

        row_index = button.pronunciation_row_index
        phrase, substitution = self._pronunciations[row_index]
        self._show_edit_dialog(phrase, substitution, row_index)

    def _on_delete_clicked(self, button: Gtk.Button) -> None:
        """Handle delete button click."""

        row_index = button.pronunciation_row_index
        phrase, _ = self._pronunciations[row_index]
        key = phrase.lower()
        if key in self._inherited_keys:
            self._deleted_inherited_keys.add(key)
            self._inherited_keys.discard(key)
        del self._pronunciations[row_index]
        self._has_unsaved_changes = True
        self.refresh()

        script = script_manager.get_manager().get_active_script()
        if script:
            presentation_manager.get_manager().present_message(
                messages.PRONUNCIATION_DELETED % phrase,
            )

    def _on_listbox_realize(self, _widget: Gtk.Widget) -> None:
        """Suppress pronunciation substitution while editing entries."""

        self._manager.suppress()

    def _on_listbox_unrealize(self, _widget: Gtk.Widget) -> None:
        """Restore pronunciation substitution when done editing entries."""

        self._manager.unsuppress()

    def _on_add_clicked(self, _button: Gtk.Button) -> None:
        """Handle Add button click to open add dialog."""

        self._show_add_dialog()

    def _show_add_dialog(self) -> None:
        """Show dialog to add a new pronunciation."""

        dialog, add_button = orca_gui_helpers.create_header_bar_dialog(
            guilabels.ADD_NEW_PRONUNCIATION,
            guilabels.DIALOG_CANCEL,
            guilabels.DIALOG_ADD,
            transient_for=self.get_toplevel(),
        )

        content_area = dialog.get_content_area()

        listbox = orca_gui_helpers.FocusManagedListBox()

        # Size group to ensure labels have same width and entries align
        label_size_group = orca_gui_helpers.create_horizontal_size_group()

        def on_entry_activate(_entry):
            """Only accept the dialog if both fields are filled."""
            if phrase_entry.get_text().strip() and substitution_entry.get_text().strip():
                dialog.response(Gtk.ResponseType.OK)

        def update_add_button_sensitivity(_entry: Gtk.Entry | None = None) -> None:
            add_button.set_sensitive(
                bool(phrase_entry.get_text().strip() and substitution_entry.get_text().strip())
            )

        phrase_entry = orca_gui_helpers.create_entry(
            size_request=(-1, 40),
            changed_handler=update_add_button_sensitivity,
            activate_handler=on_entry_activate,
        )
        phrase_row = orca_gui_helpers.create_labeled_entry_row(
            guilabels.DICTIONARY_ACTUAL_STRING,
            phrase_entry,
            include_top_separator=False,
            label_size_group=label_size_group,
        )
        listbox.add_row_with_widget(phrase_row, phrase_entry)

        substitution_entry = orca_gui_helpers.create_entry(
            size_request=(-1, 40),
            changed_handler=update_add_button_sensitivity,
            activate_handler=on_entry_activate,
        )
        substitution_row = orca_gui_helpers.create_labeled_entry_row(
            guilabels.DICTIONARY_REPLACEMENT_STRING,
            substitution_entry,
            label_size_group=label_size_group,
        )
        listbox.add_row_with_widget(substitution_row, substitution_entry)

        content_area.pack_start(listbox, True, True, 0)

        def on_response(dlg, response_id):
            if response_id == Gtk.ResponseType.OK:
                phrase = phrase_entry.get_text().strip()
                substitution = substitution_entry.get_text().strip()
                if phrase and substitution:
                    self._pronunciations.append((phrase, substitution))
                    self._has_unsaved_changes = True
                    self.refresh()
            dlg.destroy()

        dialog.connect("response", on_response)
        update_add_button_sensitivity()
        dialog.show_all()
        if add_button.get_sensitive():
            add_button.grab_default()
        phrase_entry.grab_focus()

    def _show_edit_dialog(  # pylint: disable=too-many-locals
        self,
        phrase: str,
        substitution: str,
        row_index: int,
    ) -> None:
        """Show dialog to edit an existing pronunciation."""

        dialog, edit_button = orca_gui_helpers.create_header_bar_dialog(
            guilabels.EDIT_PRONUNCIATION,
            guilabels.DIALOG_CANCEL,
            guilabels.DIALOG_EDIT,
            transient_for=self.get_toplevel(),
        )

        content_area = dialog.get_content_area()

        listbox = orca_gui_helpers.FocusManagedListBox()

        # Size group to ensure labels have same width and entries align
        label_size_group = orca_gui_helpers.create_horizontal_size_group()

        def on_entry_activate(_entry):
            """Only accept the dialog if both fields are filled."""
            if phrase_entry.get_text().strip() and substitution_entry.get_text().strip():
                dialog.response(Gtk.ResponseType.OK)

        def update_edit_button_sensitivity(_entry: Gtk.Entry | None = None) -> None:
            edit_button.set_sensitive(
                bool(phrase_entry.get_text().strip() and substitution_entry.get_text().strip())
            )

        phrase_entry = orca_gui_helpers.create_entry(
            phrase,
            size_request=(-1, 40),
            changed_handler=update_edit_button_sensitivity,
            activate_handler=on_entry_activate,
        )
        phrase_row = orca_gui_helpers.create_labeled_entry_row(
            guilabels.DICTIONARY_ACTUAL_STRING,
            phrase_entry,
            include_top_separator=False,
            label_size_group=label_size_group,
        )
        listbox.add_row_with_widget(phrase_row, phrase_entry)

        substitution_entry = orca_gui_helpers.create_entry(
            substitution,
            size_request=(-1, 40),
            changed_handler=update_edit_button_sensitivity,
            activate_handler=on_entry_activate,
        )
        substitution_row = orca_gui_helpers.create_labeled_entry_row(
            guilabels.DICTIONARY_REPLACEMENT_STRING,
            substitution_entry,
            label_size_group=label_size_group,
        )
        listbox.add_row_with_widget(substitution_row, substitution_entry)

        content_area.pack_start(listbox, True, True, 0)

        def on_response(dlg, response_id):
            if response_id == Gtk.ResponseType.OK:
                new_phrase = phrase_entry.get_text().strip()
                new_substitution = substitution_entry.get_text().strip()
                if new_phrase and new_substitution:
                    self._inherited_keys.discard(phrase.lower())
                    self._pronunciations[row_index] = (new_phrase, new_substitution)
                    self._has_unsaved_changes = True
                    self.refresh()
            dlg.destroy()

        dialog.connect("response", on_response)
        update_edit_button_sensitivity()
        dialog.show_all()
        edit_button.grab_default()
        phrase_entry.grab_focus()

    def reload(self) -> None:
        """Reload settings from the manager and refresh the UI."""

        self._pronunciations = []
        self._inherited_keys = set()
        self._deleted_inherited_keys = set()
        self._loaded_from_settings = False
        self._has_unsaved_changes = False
        self.refresh()

    def save_settings(  # pylint: disable=too-many-branches
        self,
        profile: str = "",
        app_name: str = "",
    ) -> dict[str, list[str]]:
        """Save settings and return a dictionary of the current values for those settings."""

        self._manager.set_dictionary({})
        result: dict[str, list[str]] = {}

        for phrase, substitution in self._pronunciations:
            if phrase and substitution:
                self._manager.set_pronunciation(phrase, substitution)
                result[phrase.lower()] = [phrase, substitution]

        self._has_unsaved_changes = False

        if profile:
            registry = gsettings_registry.get_registry()
            parent_pronunciations: dict[str, str] = {}
            if profile != "default" or app_name:
                if profile != "default":
                    parent_pronunciations |= registry.get_pronunciations("default", "")
                if app_name:
                    parent_pronunciations |= registry.get_pronunciations(profile, "")

            diff: dict[str, str] = {}
            for key, value in result.items():
                replacement = value[1]
                if parent_pronunciations.get(key) != replacement:
                    diff[key] = replacement
            for key in self._deleted_inherited_keys:
                if key in parent_pronunciations:
                    diff[key] = ""

            pron_gs = registry.get_settings("pronunciations", profile, "pronunciations", app_name)
            if pron_gs is not None:
                if diff:
                    pron_gs.set_value("entries", GLib.Variant("a{ss}", diff))
                elif pron_gs.get_user_value("entries") is not None:
                    pron_gs.reset("entries")

        return result

    def refresh(self) -> None:
        """Update listbox to reflect current pronunciation list."""

        if self._listbox is None:
            return

        self._cancel_pending_build()
        self._initializing = True

        for child in self._listbox.get_children():
            self._listbox.remove(child)
        self._pending_build_index = 0

        if not self._loaded_from_settings:
            self._loaded_from_settings = True
            registry = gsettings_registry.get_registry()
            profile = registry.get_active_profile()
            app_name = ""
            if self._script and self._script.app:
                from .ax_object import AXObject  # pylint: disable=import-outside-toplevel

                app_name = AXObject.get_name(self._script.app)

            pronunciation_dict = registry.layered_lookup(
                "pronunciations",
                "entries",
                "a{ss}",
                app_name=app_name or None,
                default={},
            )
            local_dict = registry.get_pronunciations(profile, app_name)
            self._inherited_keys = set(pronunciation_dict.keys()) - set(local_dict.keys())

            for key in sorted(pronunciation_dict.keys()):
                if pronunciation_dict[key]:
                    self._pronunciations.append((key, pronunciation_dict[key]))

        if self._pronunciations:
            if self.get_mapped():
                self._schedule_build_chunk()
        else:
            empty_row = orca_gui_helpers.create_info_row(guilabels.DICTIONARY_EMPTY)
            self._listbox.add(empty_row)
            empty_row.show_all()

        self._initializing = False

    def _on_map(self, _widget: Gtk.Widget) -> None:
        """Start (or resume) the chunked row build when the page becomes visible."""

        if self._pending_build_id != 0:
            return
        if self._pending_build_index < len(self._pronunciations):
            self._schedule_build_chunk()

    def _on_destroy(self, _widget: Gtk.Widget) -> None:
        """Cancel any in-flight chunked build before the widget goes away."""

        self._cancel_pending_build()
        self._listbox = None

    def _cancel_pending_build(self) -> None:
        """Stop any in-flight chunked row build."""

        if self._pending_build_id != 0:
            GLib.source_remove(self._pending_build_id)
            self._pending_build_id = 0

    def _schedule_build_chunk(self) -> None:
        """Queue an idle callback to add the next batch of rows."""

        self._pending_build_id = GLib.idle_add(self._build_chunk)

    def _build_chunk(self) -> bool:
        """Add up to BUILD_BATCH_SIZE rows. Returns True to keep idling."""

        if self._listbox is None:
            self._pending_build_id = 0
            return False

        end = min(self._pending_build_index + self.BUILD_BATCH_SIZE, len(self._pronunciations))
        for index in range(self._pending_build_index, end):
            phrase, substitution = self._pronunciations[index]
            row = self._create_row(
                phrase,
                substitution,
                index,
                include_top_separator=index > 0,
            )
            self._listbox.add(row)
            row.show_all()
        self._pending_build_index = end

        if end >= len(self._pronunciations):
            self._pending_build_id = 0
            return False
        return True



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2023 Igalia, S.L.
# Copyright 2023 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines

"""Provides Orca-controlled navigation for tabular content."""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from . import (
    command_manager,
    dbus_service,
    debug,
    focus_manager,
    gsettings_registry,
    guilabels,
    input_event,
    input_event_manager,
    messages,
    presentation_manager,
    speech_presenter,
    table_navigator_command_definitions,
)
from .ax_object import AXObject
from .ax_table import AXTable
from .ax_utilities import AXUtilities
from .ax_utilities_text import CaretSetReason
from .extension import Extension

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .command import Command
    from .input_event import InputEvent
    from .scripts import default


@gsettings_registry.get_registry().gsettings_schema(
    "org.gnome.Orca.TableNavigation",
    name="table-navigation",
)
class TableNavigator(Extension):
    """Provides Orca-controlled navigation for tabular content."""

    _SCHEMA = "table-navigation"
    KEY_ENABLED = "enabled"
    KEY_SKIP_BLANK_CELLS = "skip-blank-cells"

    def _get_setting(self, key: str, default: bool) -> bool:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            "b",
            default=default,
        )

    GROUP_LABEL = guilabels.KB_GROUP_TABLE_NAVIGATION

    def __init__(self) -> None:
        self._previous_reported_row: int | None = None
        self._previous_reported_col: int | None = None
        self._last_input_event: InputEvent | None = None
        super().__init__()

    @staticmethod
    def navigation_command(func):
        """Decorator that logs the command, records the input event, and returns True."""

        @functools.wraps(func)
        def wrapper(self, script, event=None, notify_user=True) -> bool:
            tokens = [
                "TABLE NAVIGATOR:",
                func,
                "\nScript:",
                script,
                "\nEvent:",
                event,
                "\nnotify_user:",
                notify_user,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            self.set_last_input_event(event)
            func(self, script, notify_user)
            return True

        return wrapper

    def set_last_input_event(self, event: InputEvent | None) -> None:
        """Records the input event associated with the most recent navigation command."""

        self._last_input_event = event

    def last_input_event_was_navigation_command(self) -> bool:
        """Returns true if the last input event was a navigation command."""

        if self._last_input_event is None:
            return False

        manager = input_event_manager.get_manager()
        result = manager.last_event_equals_or_is_release_for_event(self._last_input_event)
        if self._last_input_event is not None:
            string = self._last_input_event.as_single_line_string()
        else:
            string = "None"

        msg = f"TABLE NAVIGATOR: Last navigation event ({string}) is last input event: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def _get_commands(self) -> list[Command]:
        return table_navigator_command_definitions.get_commands(self)

    @dbus_service.command
    def toggle_enabled(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Toggles table navigation."""

        tokens = [
            "TABLE NAVIGATOR: toggle_enabled. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        enabled = not command_manager.get_manager().is_group_enabled(
            guilabels.KB_GROUP_TABLE_NAVIGATION,
        )

        if notify_user:
            if enabled:
                presentation_manager.get_manager().present_message(
                    messages.TABLE_NAVIGATION_ENABLED,
                )
            else:
                presentation_manager.get_manager().present_message(
                    messages.TABLE_NAVIGATION_DISABLED,
                )

        self.set_is_enabled(enabled)
        return True

    def _is_blank(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is empty or consists of only whitespace."""

        if AXUtilities.is_focusable(obj):
            tokens = ["TABLE NAVIGATOR:", obj, "is not blank: it is focusable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXObject.get_name(obj):
            tokens = ["TABLE NAVIGATOR:", obj, "is not blank: it has a name"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXObject.get_child_count(obj):
            for child in AXObject.iter_children(obj):
                if not self._is_blank(child):
                    tokens = ["TABLE NAVIGATOR:", obj, "is not blank:", child, "is not blank"]
                    debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                    return False
            return True

        if not AXUtilities.is_whitespace_or_empty(obj):
            tokens = ["TABLE NAVIGATOR:", obj, "is not blank: it has text"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["TABLE NAVIGATOR: Treating", obj, "as blank"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    def _get_current_cell(self) -> Atspi.Accessible:
        """Returns the current cell."""

        cell = focus_manager.get_manager().get_locus_of_focus()

        # We might have nested cells. So far this has only been seen in Gtk, where the
        # parent of a table cell is also a table cell. From the user's perspective, we
        # are on the parent. This check also covers Writer documents in which the caret
        # is likely in a paragraph child of the cell.
        parent = AXObject.get_parent(cell)
        if AXUtilities.is_table_cell_or_header(parent):
            cell = parent

        # And we might instead be in some deeply-nested elements which display text in
        # a web table, so we do one more check.
        if not AXUtilities.is_table_cell_or_header(cell):
            cell = AXUtilities.find_ancestor(cell, AXUtilities.is_table_cell_or_header)

        tokens = ["TABLE NAVIGATOR: Current cell is", cell]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return cell

    def _get_cell_coordinates(self, cell: Atspi.Accessible) -> tuple:
        """Returns the coordinates of cell, possibly adjusted for linear movement."""

        row, col = AXTable.get_cell_coordinates(cell, prefer_attribute=False)
        if self._previous_reported_row is None or self._previous_reported_col is None:
            return row, col

        # If we're in a cell that spans multiple rows and/or columns, the coordinates will refer to
        # the upper left cell in the spanned range(s). We're storing the last row and column that
        # we presented in order to facilitate more linear movement. Therefore, if the cell at the
        # stored coordinates is the same as cell, we prefer the stored coordinates.
        last_cell = AXTable.get_cell_at(
            AXUtilities.get_table(cell),
            self._previous_reported_row,
            self._previous_reported_col,
        )
        if last_cell == cell:
            return self._previous_reported_row, self._previous_reported_col

        return row, col

    def _vertical_cell_skipping_holes(
        self, current: Atspi.Accessible, descending: bool
    ) -> Atspi.Accessible | None:
        """Returns the nearest real cell above or below current, skipping holes in the column."""

        table = AXUtilities.get_table(current)
        if table is None:
            return None

        row, col = AXTable.get_cell_coordinates(current, prefer_attribute=False)
        if descending:
            row += AXTable.get_cell_spans(current, prefer_attribute=False)[0]
            stop, step = AXTable.get_row_count(table, prefer_attribute=False), 1
        else:
            row -= 1
            stop, step = -1, -1

        while row != stop:
            if (cell := AXTable.get_cell_at(table, row, col)) is not None:
                return cell
            row += step

        return None

    def _horizontal_cell_skipping_holes(
        self, current: Atspi.Accessible, rightward: bool
    ) -> Atspi.Accessible | None:
        """Returns the nearest real cell beside current, skipping holes in its own row."""

        table = AXUtilities.get_table(current)
        if table is None:
            return None

        row, col = AXTable.get_cell_coordinates(current, prefer_attribute=False)
        row_span, col_span = AXTable.get_cell_spans(current, prefer_attribute=False)
        col = col + col_span if rightward else col - 1
        last_row = row + row_span

        # A cell that spans rows but sits alone in its own row has only holes beside it; its real
        # neighbors live in the lower rows its rowspan covers.
        while row < last_row:
            if (cell := AXTable.get_cell_at(table, row, col)) is not None:
                return cell
            row += 1

        return None

    @dbus_service.command
    @navigation_command
    def move_left(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the cell on the left."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_start_of_row(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_ROW_BEGINNING)
            return

        row, col = self._get_cell_coordinates(current)
        cell = AXTable.get_cell_on_left(current)

        if self.get_skip_blank_cells():
            while cell and self._is_blank(cell) and not AXTable.is_start_of_row(cell):
                cell = AXTable.get_cell_on_left(cell)

        if cell is None:
            cell = self._horizontal_cell_skipping_holes(current, rightward=False)

        if cell is not None:
            row, col = AXTable.get_cell_coordinates(cell, prefer_attribute=False)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_right(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the cell on the right."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_end_of_row(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_ROW_END)
            return

        row, col = self._get_cell_coordinates(current)
        cell = AXTable.get_cell_on_right(current)

        if self.get_skip_blank_cells():
            while cell and self._is_blank(cell) and not AXTable.is_end_of_row(cell):
                cell = AXTable.get_cell_on_right(cell)

        if cell is None:
            cell = self._horizontal_cell_skipping_holes(current, rightward=True)

        if cell is not None:
            row, col = AXTable.get_cell_coordinates(cell, prefer_attribute=False)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_up(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the cell above."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_top_of_column(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_COLUMN_TOP)
            return

        row, col = self._get_cell_coordinates(current)
        cell = AXTable.get_cell_above(current)

        if self.get_skip_blank_cells():
            while cell and self._is_blank(cell) and not AXTable.is_top_of_column(cell):
                cell = AXTable.get_cell_above(cell)

        if cell is None:
            cell = self._vertical_cell_skipping_holes(current, descending=False)

        if cell is not None:
            row, col = AXTable.get_cell_coordinates(cell, prefer_attribute=False)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_down(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the cell below."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_bottom_of_column(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_COLUMN_BOTTOM)
            return

        row, col = self._get_cell_coordinates(current)
        cell = AXTable.get_cell_below(current)

        if self.get_skip_blank_cells():
            while cell and self._is_blank(cell) and not AXTable.is_bottom_of_column(cell):
                cell = AXTable.get_cell_below(cell)

        if cell is None:
            cell = self._vertical_cell_skipping_holes(current, descending=True)

        if cell is not None:
            row, col = AXTable.get_cell_coordinates(cell, prefer_attribute=False)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_to_first_cell(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the first cell."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        table = AXUtilities.get_table(current)
        cell = AXTable.get_first_cell(table)
        self._present_cell(script, cell, 0, 0, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_to_last_cell(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the last cell."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        table = AXUtilities.get_table(current)
        cell = AXTable.get_last_cell(table)
        self._present_cell(
            script,
            cell,
            AXTable.get_row_count(table),
            AXTable.get_column_count(table),
            current,
            notify_user,
        )

    @dbus_service.command
    @navigation_command
    def move_to_beginning_of_row(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the beginning of the row."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_start_of_row(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_ROW_BEGINNING)
            return

        cell = AXTable.get_start_of_row(current)
        row, col = self._get_cell_coordinates(cell)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_to_end_of_row(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the end of the row."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_end_of_row(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_ROW_END)
            return

        cell = AXTable.get_end_of_row(current)
        row, col = self._get_cell_coordinates(cell)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_to_top_of_column(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the top of the column."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_top_of_column(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_COLUMN_TOP)
            return

        cell = AXTable.get_top_of_column(current)
        row, col = self._get_cell_coordinates(cell)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def move_to_bottom_of_column(self, script: default.Script, notify_user: bool = True) -> None:
        """Moves to the bottom of the column."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        if AXTable.is_bottom_of_column(current):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_COLUMN_BOTTOM)
            return

        cell = AXTable.get_bottom_of_column(current)
        row, col = self._get_cell_coordinates(cell)
        self._present_cell(script, cell, row, col, current, notify_user)

    @dbus_service.command
    @navigation_command
    def set_dynamic_column_headers_row(
        self, _script: default.Script, notify_user: bool = True
    ) -> None:
        """Sets the row for the dynamic header columns to the current row."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        table = AXUtilities.get_table(current)
        if table:
            row = AXTable.get_cell_coordinates(current)[0]
            AXUtilities.set_dynamic_column_headers_row(table, row)
            if notify_user:
                presentation_manager.get_manager().present_message(
                    messages.DYNAMIC_COLUMN_HEADER_SET % (row + 1),
                )

    @dbus_service.command
    @navigation_command
    def clear_dynamic_column_headers_row(
        self, _script: default.Script, notify_user: bool = True
    ) -> None:
        """Clears the row for the dynamic column headers."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        table = AXUtilities.get_table(focus_manager.get_manager().get_locus_of_focus())
        if table:
            AXUtilities.clear_dynamic_column_headers_row(table)
            if notify_user:
                presentation_manager.get_manager().interrupt_presentation()
                presentation_manager.get_manager().present_message(
                    messages.DYNAMIC_COLUMN_HEADER_CLEARED,
                )

    @dbus_service.command
    @navigation_command
    def set_dynamic_row_headers_column(
        self, _script: default.Script, notify_user: bool = True
    ) -> None:
        """Sets the column for the dynamic row headers to the current column."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        table = AXUtilities.get_table(current)
        if table:
            column = AXTable.get_cell_coordinates(current)[1]
            AXUtilities.set_dynamic_row_headers_column(table, column)
            if notify_user:
                presentation_manager.get_manager().present_message(
                    messages.DYNAMIC_ROW_HEADER_SET % AXUtilities.get_column_label(table, column),
                )

    @dbus_service.command
    @navigation_command
    def clear_dynamic_row_headers_column(
        self, _script: default.Script, notify_user: bool = True
    ) -> None:
        """Clears the column for the dynamic row headers."""

        current = self._get_current_cell()
        if current is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return

        table = AXUtilities.get_table(focus_manager.get_manager().get_locus_of_focus())
        if table:
            AXUtilities.clear_dynamic_row_headers_column(table)
            if notify_user:
                presentation_manager.get_manager().interrupt_presentation()
                presentation_manager.get_manager().present_message(
                    messages.DYNAMIC_ROW_HEADER_CLEARED,
                )

    def _present_cell(
        self,
        script: default.Script,
        cell: Atspi.Accessible,
        row: int,
        col: int,
        previous_cell: Atspi.Accessible,
        notify_user: bool = True,
    ) -> None:
        """Presents cell to the user."""

        if not AXUtilities.is_table_cell_or_header(cell):
            tokens = ["TABLE NAVIGATOR: ", cell, f"(row {row}, column {col}) is not cell or header"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return

        self._previous_reported_row = row
        self._previous_reported_col = col

        if script.utilities.grab_focus_when_setting_caret(cell):
            AXObject.grab_focus(cell)

        obj = AXUtilities.get_descendant_supporting_text(cell) or cell
        focus_mgr = focus_manager.get_manager()
        focus_mgr.set_locus_of_focus(None, obj, False)
        focus_mgr.emit_region_changed(obj, mode=focus_manager.TABLE_NAVIGATOR)

        if AXObject.supports_text(obj) and not AXUtilities.is_gui_cell(cell):
            script.utilities.set_caret_position(obj, 0, reason=CaretSetReason.TABLE_NAVIGATION)

        if not notify_user:
            msg = "TABLE NAVIGATOR: _present_cell called with notify_user=False"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
        script.present_object(cell, offset=0, prior_obj=previous_cell)

        manager = speech_presenter.get_presenter()
        # TODO - JD: This should be part of the normal table cell presentation.
        if manager.get_announce_cell_coordinates():
            presentation_manager.get_manager().present_message(
                messages.TABLE_CELL_COORDINATES % {"row": row + 1, "column": col + 1},
            )

        # TODO - JD: Ditto.
        if manager.get_announce_cell_span():
            rowspan, colspan = AXTable.get_cell_spans(cell)
            if rowspan > 1 or colspan > 1:
                presentation_manager.get_manager().present_message(
                    messages.cell_span(rowspan, colspan),
                )

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ENABLED,
        schema="table-navigation",
        gtype="b",
        default=True,
        summary="Enable table navigation",
        migration_key="tableNavigationEnabled",
    )
    @dbus_service.getter
    def get_is_enabled(self) -> bool:
        """Returns whether table navigation is enabled."""

        return self._get_setting(self.KEY_ENABLED, True)

    @dbus_service.setter
    def set_is_enabled(self, value: bool) -> bool:
        """Sets whether table navigation is enabled."""

        if self.get_is_enabled() == value:
            msg = f"TABLE NAVIGATOR: Enabled already {value}. Refreshing command group."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            command_manager.get_manager().set_group_enabled(
                guilabels.KB_GROUP_TABLE_NAVIGATION,
                value,
            )
            return True

        msg = f"TABLE NAVIGATOR: Setting enabled to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(self._SCHEMA, self.KEY_ENABLED, value)

        self._last_input_event = None
        command_manager.get_manager().set_group_enabled(guilabels.KB_GROUP_TABLE_NAVIGATION, value)

        return True

    def refresh_enabled_state(self) -> None:
        """Re-applies the enabled state for the currently active app."""

        enabled = self.get_is_enabled()
        msg = f"TABLE NAVIGATOR: Refreshing enabled state for active app: {enabled}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        command_manager.get_manager().set_group_enabled(
            guilabels.KB_GROUP_TABLE_NAVIGATION,
            enabled,
        )

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SKIP_BLANK_CELLS,
        schema="table-navigation",
        gtype="b",
        default=False,
        summary="Skip blank cells during navigation",
        migration_key="skipBlankCells",
    )
    @dbus_service.getter
    def get_skip_blank_cells(self) -> bool:
        """Returns whether blank cells should be skipped during navigation."""

        return self._get_setting(self.KEY_SKIP_BLANK_CELLS, False)

    @dbus_service.setter
    def set_skip_blank_cells(self, value: bool) -> bool:
        """Sets whether blank cells should be skipped during navigation."""

        if self.get_skip_blank_cells() == value:
            return True

        msg = f"TABLE NAVIGATOR: Setting skip blank cells to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_SKIP_BLANK_CELLS, value
        )
        return True


_navigator: TableNavigator = TableNavigator()


def get_navigator() -> TableNavigator:
    """Returns the Table Navigator"""

    return _navigator



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for Say All support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import guilabels, preferences_grid_base, say_all_presenter, text_attribute_manager

if TYPE_CHECKING:
    from .say_all_presenter import SayAllPresenter


class SayAllPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Say All preferences page."""

    _gsettings_schema = "say-all"
    _documentation_summary = "Use these settings to control Orca's continuous reading command."

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for Say All preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.GENERAL_SAY_ALL,
            panel_id="say_all_presenter.say-all",
            description="Say All settings control Orca's continuous reading command.",
            schema="say-all",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SAY_ALL_BY,
                    kind="choice",
                    summary=(
                        "Say All reads from the current location to the end of the document; "
                        "this controls whether it pauses after each sentence or after each line."
                    ),
                    schema="say-all",
                    key=say_all_presenter.SayAllPresenter.KEY_STYLE,
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.SAY_ALL_STYLE_SENTENCE,
                            value=say_all_presenter.SayAllStyle.SENTENCE.string_name,
                            summary="Pauses after each sentence.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.SAY_ALL_STYLE_LINE,
                            value=say_all_presenter.SayAllStyle.LINE.string_name,
                            summary="Pauses after each line.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPEECH_ONLY_SPEAK_DISPLAYED_TEXT,
                    kind="switch",
                    summary=("Controls whether Say All skips text that is not visually displayed."),
                    schema="say-all",
                    key=say_all_presenter.SayAllPresenter.KEY_ONLY_SPEAK_DISPLAYED_TEXT,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SAY_ALL_REWIND_AND_FAST_FORWARD_BY,
                    kind="group",
                    summary=(
                        "Controls which commands can move within the document while Say All "
                        "is active."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.SAY_ALL_UP_AND_DOWN_ARROW,
                            kind="switch",
                            summary=(
                                "Allows Up Arrow and Down Arrow to move within the document "
                                "during Say All so you can re-hear recent content or skip "
                                "ahead without restarting Say All."
                            ),
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_REWIND_AND_FAST_FORWARD,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.SAY_ALL_STRUCTURAL_NAVIGATION,
                            kind="switch",
                            summary=(
                                "Allows structural navigation commands, such as heading and "
                                "paragraph navigation, to move backward and forward during "
                                "Say All."
                            ),
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_STRUCTURAL_NAVIGATION,
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.CHANGE_ANNOUNCEMENTS,
                    kind="group",
                    summary=(
                        "Controls which document revision and text formatting changes Orca "
                        "announces during Say All."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_TRACKED_CHANGES,
                            kind="switch",
                            summary=(
                                "Controls whether Orca announces tracked changes during Say All."
                            ),
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_TRACKED_CHANGES,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.TEXT_ATTRIBUTE_CHANGES,
                            kind="choice",
                            summary=(
                                "Controls when Orca announces text formatting changes during "
                                "Say All."
                            ),
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_TEXT_ATTRIBUTE_CHANGE_MODE,
                            value_docs=(
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.TEXT_ATTRIBUTE_CHANGES_OFF,
                                    value="off",
                                    summary="Does not announce formatting changes.",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.TEXT_ATTRIBUTE_CHANGES_EDITABLE,
                                    value="editable-only",
                                    summary="Announces formatting changes only in editable text.",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.TEXT_ATTRIBUTE_CHANGES_ALWAYS,
                                    value="always",
                                    summary="Always announces formatting changes.",
                                ),
                            ),
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.ANNOUNCEMENTS,
                    kind="group",
                    summary=(
                        "Controls which document and web-container boundaries Orca announces "
                        "during Say All."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_ARTICLES,
                            kind="switch",
                            summary="Controls whether Orca announces articles during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_ARTICLE,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_BLOCKQUOTES,
                            kind="switch",
                            summary="Controls whether Orca announces blockquotes during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_BLOCKQUOTE,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_CODE_BLOCKS,
                            kind="switch",
                            summary="Controls whether Orca announces code blocks during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_CODE_BLOCK,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_EMBEDDED_DOCUMENTS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca announces embedded documents during Say All."
                            ),
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_DOCUMENT,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_FORMS,
                            kind="switch",
                            summary="Controls whether Orca announces forms during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_FORM,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_LANDMARKS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca announces ARIA landmarks during Say All."
                            ),
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_LANDMARK,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_LISTS,
                            kind="switch",
                            summary="Controls whether Orca announces lists during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_LIST,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_PANELS,
                            kind="switch",
                            summary="Controls whether Orca announces panels during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_GROUPING,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ANNOUNCE_TABLES,
                            kind="switch",
                            summary="Controls whether Orca announces tables during Say All.",
                            schema="say-all",
                            key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_TABLE,
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: SayAllPresenter) -> None:
        self._only_speak_displayed_control = preferences_grid_base.BooleanPreferenceControl(
            label=guilabels.SPEECH_ONLY_SPEAK_DISPLAYED_TEXT,
            getter=presenter.get_only_speak_displayed_text,
            setter=presenter.set_only_speak_displayed_text,
            prefs_key=say_all_presenter.SayAllPresenter.KEY_ONLY_SPEAK_DISPLAYED_TEXT,
        )

        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.SAY_ALL_BY,
                options=[guilabels.SAY_ALL_STYLE_SENTENCE, guilabels.SAY_ALL_STYLE_LINE],
                values=[
                    say_all_presenter.SayAllStyle.SENTENCE.value,
                    say_all_presenter.SayAllStyle.LINE.value,
                ],
                getter=presenter.get_style_as_int,
                setter=presenter.set_style_from_int,
                prefs_key=say_all_presenter.SayAllPresenter.KEY_STYLE,
            ),
            self._only_speak_displayed_control,
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SAY_ALL_UP_AND_DOWN_ARROW,
                getter=presenter.get_rewind_and_fast_forward_enabled,
                setter=presenter.set_rewind_and_fast_forward_enabled,
                prefs_key=say_all_presenter.SayAllPresenter.KEY_REWIND_AND_FAST_FORWARD,
                member_of=guilabels.SAY_ALL_REWIND_AND_FAST_FORWARD_BY,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SAY_ALL_STRUCTURAL_NAVIGATION,
                getter=presenter.get_structural_navigation_enabled,
                setter=presenter.set_structural_navigation_enabled,
                prefs_key=say_all_presenter.SayAllPresenter.KEY_STRUCTURAL_NAVIGATION,
                member_of=guilabels.SAY_ALL_REWIND_AND_FAST_FORWARD_BY,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.ANNOUNCE_TRACKED_CHANGES,
                getter=presenter.get_announce_tracked_changes,
                setter=presenter.set_announce_tracked_changes,
                prefs_key=say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_TRACKED_CHANGES,
                member_of=guilabels.CHANGE_ANNOUNCEMENTS,
                determine_sensitivity=self._only_speak_displayed_text_is_off,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.TEXT_ATTRIBUTE_CHANGES,
                options=[
                    guilabels.TEXT_ATTRIBUTE_CHANGES_OFF,
                    guilabels.TEXT_ATTRIBUTE_CHANGES_EDITABLE,
                    guilabels.TEXT_ATTRIBUTE_CHANGES_ALWAYS,
                ],
                values=[
                    text_attribute_manager.TextAttributeChangeMode.OFF.value,
                    text_attribute_manager.TextAttributeChangeMode.EDITABLE_ONLY.value,
                    text_attribute_manager.TextAttributeChangeMode.ALWAYS.value,
                ],
                getter=presenter.get_text_attribute_change_mode_as_int,
                setter=presenter.set_text_attribute_change_mode_from_int,
                prefs_key=say_all_presenter.SayAllPresenter.KEY_TEXT_ATTRIBUTE_CHANGE_MODE,
                member_of=guilabels.CHANGE_ANNOUNCEMENTS,
                determine_sensitivity=self._only_speak_displayed_text_is_off,
            ),
            *[
                preferences_grid_base.BooleanPreferenceControl(
                    label=label,
                    getter=getter,
                    setter=setter,
                    prefs_key=key,
                    member_of=guilabels.ANNOUNCEMENTS,
                    determine_sensitivity=self._only_speak_displayed_text_is_off,
                )
                for label, getter, setter, key in [
                    (
                        guilabels.ANNOUNCE_ARTICLES,
                        presenter.get_announce_article,
                        presenter.set_announce_article,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_ARTICLE,
                    ),
                    (
                        guilabels.ANNOUNCE_BLOCKQUOTES,
                        presenter.get_announce_blockquote,
                        presenter.set_announce_blockquote,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_BLOCKQUOTE,
                    ),
                    (
                        guilabels.ANNOUNCE_CODE_BLOCKS,
                        presenter.get_announce_code_block,
                        presenter.set_announce_code_block,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_CODE_BLOCK,
                    ),
                    (
                        guilabels.ANNOUNCE_EMBEDDED_DOCUMENTS,
                        presenter.get_announce_document,
                        presenter.set_announce_document,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_DOCUMENT,
                    ),
                    (
                        guilabels.ANNOUNCE_FORMS,
                        presenter.get_announce_form,
                        presenter.set_announce_form,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_FORM,
                    ),
                    (
                        guilabels.ANNOUNCE_LANDMARKS,
                        presenter.get_announce_landmark,
                        presenter.set_announce_landmark,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_LANDMARK,
                    ),
                    (
                        guilabels.ANNOUNCE_LISTS,
                        presenter.get_announce_list,
                        presenter.set_announce_list,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_LIST,
                    ),
                    (
                        guilabels.ANNOUNCE_PANELS,
                        presenter.get_announce_grouping,
                        presenter.set_announce_grouping,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_GROUPING,
                    ),
                    (
                        guilabels.ANNOUNCE_TABLES,
                        presenter.get_announce_table,
                        presenter.set_announce_table,
                        say_all_presenter.SayAllPresenter.KEY_ANNOUNCE_TABLE,
                    ),
                ]
            ],
        ]

        info = (
            f"{guilabels.SAY_ALL_INFO}\n\n{guilabels.SAY_ALL_NAVIGATION_INFO}"
            f"\n\n{guilabels.SAY_ALL_ANNOUNCEMENTS_INFO}"
        )
        super().__init__(guilabels.GENERAL_SAY_ALL, controls, info_message=info)

    def _only_speak_displayed_text_is_off(self) -> bool:
        """Returns True if only-speak-displayed-text is off in the UI."""

        return not self._get_control_value(self._only_speak_displayed_control, False)



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

"""Preferences grids for speech voices and voice types."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import GObject, Gtk

from . import (
    gsettings_registry,
    guilabels,
    language_utilities,
    orca_gui_helpers,
    preferences_grid_base,
    speechserver,
)
from .acss import ACSS
from .speechserver import CapitalizationStyle, PunctuationStyle

if TYPE_CHECKING:
    from .speech_manager import SpeechManager


# pylint: disable-next=too-many-instance-attributes
class VoicesPreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """GtkGrid containing the Voice settings page."""

    _VOICE_SCHEMA = "voice"

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for speech output preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.VOICE_GLOBAL_VOICE_SETTINGS,
            panel_id="manual.global-voice-settings",
            summary="Use these settings to enable speech, select the speech backend, and "
            "configure global speech behavior.",
            description=(
                "Global voice settings apply to Orca's spoken output overall, including "
                "which speech system and synthesizer are used and how numbers, "
                "punctuation, capitalization, and languages are handled."
            ),
            schema="speech",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VOICE_SPEECH_SYSTEM,
                    kind="choice",
                    summary="Selects the speech system Orca uses. Supported systems include "
                    "Speech Dispatcher and Spiel.",
                    schema="speech",
                    key="speech-server",
                    dynamic_values=True,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VOICE_SPEECH_SYNTHESIZER,
                    kind="choice",
                    summary="Selects the speech synthesizer. Examples include espeak-ng and Voxin.",
                    schema="speech",
                    key="synthesizer",
                    dynamic_values=True,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.PUNCTUATION_STYLE,
                    kind="choice",
                    summary="Controls how much punctuation Orca speaks.",
                    schema="speech",
                    key="punctuation-level",
                    values=(
                        guilabels.PUNCTUATION_STYLE_NONE,
                        guilabels.PUNCTUATION_STYLE_SOME,
                        guilabels.PUNCTUATION_STYLE_MOST,
                        guilabels.PUNCTUATION_STYLE_ALL,
                    ),
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PUNCTUATION_STYLE_NONE,
                            summary="Requests that punctuation symbols not be spoken.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PUNCTUATION_STYLE_SOME,
                            summary=(
                                "Requests that some punctuation symbols be spoken. The "
                                "speech system determines which symbols are included."
                            ),
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PUNCTUATION_STYLE_MOST,
                            summary=(
                                "Requests that most punctuation symbols be spoken. The "
                                "speech system determines which symbols are included."
                            ),
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PUNCTUATION_STYLE_ALL,
                            summary="Requests that all punctuation symbols be spoken.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VOICE_CAPITALIZATION_STYLE,
                    kind="choice",
                    summary="Controls how Orca indicates uppercase letters.",
                    schema="speech",
                    key="capitalization-style",
                    values=(
                        guilabels.CAPITALIZATION_STYLE_VOICE_ONLY,
                        guilabels.CAPITALIZATION_STYLE_ICON,
                        guilabels.CAPITALIZATION_STYLE_SPELL,
                    ),
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.CAPITALIZATION_STYLE_VOICE_ONLY,
                            summary="Uses the configured uppercase voice without adding "
                            "another indication.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.CAPITALIZATION_STYLE_ICON,
                            summary=(
                                "Uses the configured uppercase voice and plays a tone "
                                "before speaking an uppercase letter."
                            ),
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.CAPITALIZATION_STYLE_SPELL,
                            summary=(
                                "Uses the configured uppercase voice and says capital "
                                "before speaking an uppercase letter."
                            ),
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VOICE_SPEAK_NUMBERS_AS_DIGITS,
                    kind="switch",
                    summary="When enabled, Orca speaks 123 as 1 2 3. When disabled, "
                    "Orca sends the number to the synthesizer, which likely speaks it "
                    "as one hundred and twenty three.",
                    schema="speech",
                    key="speak-numbers-as-digits",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPEECH_SPEAK_COLORS_AS_NAMES,
                    kind="switch",
                    summary="Controls whether Orca speaks colors by name, such as light "
                    "blue, instead of as RGB values.",
                    schema="speech",
                    key="use-color-names",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPEECH_BREAK_INTO_CHUNKS,
                    kind="switch",
                    summary="Controls whether Orca inserts brief pauses between parts "
                    "of a spoken presentation, such as name, role, state, mnemonic, "
                    "and tutorial message.",
                    schema="speech",
                    key="insert-pauses-between-utterances",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPEECH_USE_PRONUNCIATION_DICTIONARY,
                    kind="switch",
                    schema="speech",
                    key="use-pronunciation-dictionary",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.LANGUAGE_SWITCHING,
                    kind="group",
                    summary="Controls when Orca switches voices based on language.",
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.AUTO_LANGUAGE_SWITCHING,
                            kind="switch",
                            schema="speech",
                            key="auto-language-switching",
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.AUTO_LANGUAGE_SWITCHING_UI,
                            kind="switch",
                            schema="speech",
                            key="auto-language-switching-ui",
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.ONLY_SWITCH_CONFIGURED_LANGUAGES,
                            kind="switch",
                            schema="speech",
                            key="only-switch-configured-languages",
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, manager: SpeechManager, app_name: str = "") -> None:
        super().__init__(guilabels.SPEECH)
        self._manager = manager
        self._app_name: str = app_name
        self._initializing = True

        self._voices: dict[str, ACSS] = {}
        for vt in speechserver.VoiceType:
            self._voices[vt] = manager.get_voice_properties(vt, app_name=self._app_name)

        self._voice_families: list[speechserver.VoiceFamily] = []

        self._speech_systems_combo: Gtk.ComboBox
        self._speech_synthesizers_combo: Gtk.ComboBox
        self._punctuation_combo: Gtk.ComboBox
        self._capitalization_combo: Gtk.ComboBox
        self._global_section: Gtk.Box | None = None
        self._voice_types_section: Gtk.Box | None = None

        self._families_sorted: bool = False

        self._build()
        self._populate_speech_systems()
        self.refresh()

    def _build(self) -> None:
        """Create the Gtk widgets composing the grid."""

        row = 0

        self._global_section, global_content, global_heading = orca_gui_helpers.create_section_box(
            guilabels.VOICE_GLOBAL_VOICE_SETTINGS,
            margin_top=12,
        )

        punctuation_model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_INT)
        punctuation_model.append([guilabels.PUNCTUATION_STYLE_NONE, PunctuationStyle.NONE.value])
        punctuation_model.append([guilabels.PUNCTUATION_STYLE_SOME, PunctuationStyle.SOME.value])
        punctuation_model.append([guilabels.PUNCTUATION_STYLE_MOST, PunctuationStyle.MOST.value])
        punctuation_model.append([guilabels.PUNCTUATION_STYLE_ALL, PunctuationStyle.ALL.value])

        capitalization_model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        capitalization_model.append(
            [guilabels.CAPITALIZATION_STYLE_VOICE_ONLY, CapitalizationStyle.NONE.value],
        )
        capitalization_model.append(
            [guilabels.CAPITALIZATION_STYLE_ICON, CapitalizationStyle.ICON.value],
        )
        capitalization_model.append(
            [guilabels.CAPITALIZATION_STYLE_SPELL, CapitalizationStyle.SPELL.value],
        )

        global_listbox = preferences_grid_base.PreferencesFocusManagedListBox()
        orca_gui_helpers.add_labelled_by(global_listbox, global_heading)
        combo_size_group = orca_gui_helpers.create_horizontal_size_group()

        row_data = [
            (
                guilabels.VOICE_SPEECH_SYSTEM,
                Gtk.ListStore(GObject.TYPE_STRING),
                self._on_speech_system_changed,
            ),
            (
                guilabels.VOICE_SPEECH_SYNTHESIZER,
                Gtk.ListStore(GObject.TYPE_STRING),
                self._on_speech_synthesizer_changed,
            ),
            (guilabels.PUNCTUATION_STYLE, punctuation_model, self._on_punctuation_changed),
            (
                guilabels.VOICE_CAPITALIZATION_STYLE,
                capitalization_model,
                self._on_capitalization_changed,
            ),
        ]

        global_combos = []
        for label_text, model, changed_handler in row_data:
            row_widget, combo, _label = orca_gui_helpers.create_combo_box_row(
                label_text,
                model,
                changed_handler,
                include_top_separator=False,
            )
            combo_size_group.add_widget(combo)
            global_listbox.add_row_with_widget(row_widget, combo)
            global_combos.append(combo)

        self._speech_systems_combo = global_combos[0]
        self._speech_synthesizers_combo = global_combos[1]
        self._punctuation_combo = global_combos[2]
        self._capitalization_combo = global_combos[3]

        switch_data = [
            (
                guilabels.VOICE_SPEAK_NUMBERS_AS_DIGITS,
                self._on_speak_numbers_toggled,
                self._manager.get_speak_numbers_as_digits(),
            ),
            (
                guilabels.SPEECH_SPEAK_COLORS_AS_NAMES,
                self._on_use_color_names_toggled,
                self._manager.get_use_color_names(),
            ),
            (
                guilabels.SPEECH_BREAK_INTO_CHUNKS,
                self._on_enable_pause_breaks_toggled,
                self._manager.get_insert_pauses_between_utterances(),
            ),
            (
                guilabels.SPEECH_USE_PRONUNCIATION_DICTIONARY,
                self._on_use_pronunciation_dict_toggled,
                self._manager.get_use_pronunciation_dictionary(),
            ),
        ]

        switches = []
        for label_text, handler, state in switch_data:
            row_widget, switch, _label = orca_gui_helpers.create_switch_row(
                label_text,
                handler,
                state,
                include_top_separator=False,
            )
            global_listbox.add_row_with_widget(row_widget, switch)
            switches.append(switch)

        self._speak_numbers_switch = switches[0]
        self._use_color_names_switch = switches[1]
        self._enable_pause_breaks_switch = switches[2]
        self._use_pronunciation_dict_switch = switches[3]

        orca_gui_helpers.add_section_content(self._global_section, global_content, global_listbox)
        self.attach(self._global_section, 0, row, 1, 1)
        row += 1

        lang_switch_section, lang_switch_content, lang_switch_heading = (
            orca_gui_helpers.create_section_box(
                guilabels.LANGUAGE_SWITCHING,
                margin_top=12,
            )
        )

        lang_switch_data = [
            (
                guilabels.AUTO_LANGUAGE_SWITCHING,
                self._on_auto_language_switching_content_toggled,
                self._manager.get_auto_language_switching(),
            ),
            (
                guilabels.AUTO_LANGUAGE_SWITCHING_UI,
                self._on_auto_language_switching_ui_toggled,
                self._manager.get_auto_language_switching_ui(),
            ),
            (
                guilabels.ONLY_SWITCH_CONFIGURED_LANGUAGES,
                self._on_only_switch_configured_languages_toggled,
                self._manager.get_only_switch_configured_languages(),
            ),
        ]

        lang_switch_listbox = preferences_grid_base.PreferencesFocusManagedListBox()
        orca_gui_helpers.add_labelled_by(lang_switch_listbox, lang_switch_heading)
        lang_switches = []
        for label_text, handler, state in lang_switch_data:
            row_widget, switch, _label = orca_gui_helpers.create_switch_row(
                label_text,
                handler,
                state,
                include_top_separator=False,
            )
            lang_switch_listbox.add_row_with_widget(row_widget, switch)
            lang_switches.append(switch)

        self._auto_language_switching_content_switch = lang_switches[0]
        self._auto_language_switching_ui_switch = lang_switches[1]
        self._only_switch_configured_languages_switch = lang_switches[2]

        orca_gui_helpers.add_section_content(
            lang_switch_section,
            lang_switch_content,
            lang_switch_listbox,
        )
        self.attach(lang_switch_section, 0, row, 1, 1)

        self.show_all()  # pylint: disable=no-member

    # TODO - JD: Remove this function if it continues to prove unnecessary
    # pylint: disable-next=useless-parent-delegation
    def has_changes(self) -> bool:
        """Return True if there are unsaved changes."""

        return super().has_changes()

    def reload(self) -> None:
        """Reload settings from manager and refresh the UI."""

        app = self._app_name
        for vt in speechserver.VoiceType:
            self._voices[vt] = self._manager.get_voice_properties(vt, app_name=app)

        self._voice_families = self._manager.get_voice_families()
        self._families_sorted = False

        self._has_unsaved_changes = False
        self.refresh()

    def save_settings(self) -> dict[str, dict | list | int | str | bool]:
        """Save settings and return a dictionary of the current values for those settings."""

        result: dict[str, dict | list | int | str | bool] = {
            "voices": {vt: dict(acss) for vt, acss in self._voices.items()},
        }

        result[self._manager.KEY_SPEECH_SERVER] = self._manager.get_speech_server()
        result[self._manager.KEY_SYNTHESIZER] = self._manager.get_synthesizer()
        result[self._manager.KEY_SPEECH_SERVER_FACTORY] = self._manager.get_speech_server_factory()
        result[self._manager.KEY_PUNCTUATION_LEVEL] = self._manager.get_punctuation_level()
        result[self._manager.KEY_CAPITALIZATION_STYLE] = self._manager.get_capitalization_style()

        result[self._manager.KEY_SPEAK_NUMBERS_AS_DIGITS] = self._speak_numbers_switch.get_active()
        result[self._manager.KEY_USE_COLOR_NAMES] = self._use_color_names_switch.get_active()
        result[self._manager.KEY_INSERT_PAUSES_BETWEEN_UTTERANCES] = (
            self._enable_pause_breaks_switch.get_active()
        )
        result[self._manager.KEY_USE_PRONUNCIATION_DICTIONARY] = (
            self._use_pronunciation_dict_switch.get_active()
        )
        result[self._manager.KEY_AUTO_LANGUAGE_SWITCHING] = (
            self._auto_language_switching_content_switch.get_active()
        )
        result[self._manager.KEY_AUTO_LANGUAGE_SWITCHING_UI] = (
            self._auto_language_switching_ui_switch.get_active()
        )
        result[self._manager.KEY_ONLY_SWITCH_CONFIGURED_LANGUAGES] = (
            self._only_switch_configured_languages_switch.get_active()
        )

        self._has_unsaved_changes = False
        return result

    def refresh(self) -> None:
        """Update widget states to reflect current settings."""

        self._initializing = True

        self._populate_speech_systems()
        self._initializing = True

        app = self._app_name
        model = self._punctuation_combo.get_model()
        if model:
            current_level = PunctuationStyle[
                self._manager.get_punctuation_level(app_name=app).upper()
            ].value
            for i, row in enumerate(model):
                if row[1] == current_level:
                    self._punctuation_combo.set_active(i)
                    break

        model = self._capitalization_combo.get_model()
        if model:
            for i, row in enumerate(model):
                if row[1] == self._manager.get_capitalization_style(app_name=app):
                    self._capitalization_combo.set_active(i)
                    break

        self._speak_numbers_switch.set_active(
            self._manager.get_speak_numbers_as_digits(app_name=app),
        )
        self._use_color_names_switch.set_active(
            self._manager.get_use_color_names(app_name=app),
        )
        self._enable_pause_breaks_switch.set_active(
            self._manager.get_insert_pauses_between_utterances(app_name=app),
        )
        self._use_pronunciation_dict_switch.set_active(
            self._manager.get_use_pronunciation_dictionary(app_name=app),
        )
        self._auto_language_switching_content_switch.set_active(
            self._manager.get_auto_language_switching(app_name=app),
        )
        self._auto_language_switching_ui_switch.set_active(
            self._manager.get_auto_language_switching_ui(app_name=app),
        )
        self._only_switch_configured_languages_switch.set_active(
            self._manager.get_only_switch_configured_languages(app_name=app),
        )

        # Note: Voice type widgets are created on-demand in dialogs, so no need to refresh them here

        self._initializing = False

    def _get_acss_for_voice_type(self, voice_type: str) -> ACSS:
        """Return the local ACSS copy for the given voice type."""

        return self._voices.get(voice_type, self._voices[speechserver.VoiceType.DEFAULT])

    def _populate_speech_systems(self) -> None:
        """Populate the speech systems combo."""

        self._initializing = True

        model = self._speech_systems_combo.get_model()
        if not model:
            model = Gtk.ListStore(str)
        self._speech_systems_combo.set_model(None)
        model.clear()

        available = self._manager.get_available_servers()
        for server_name in available:
            model.append([server_name])

        self._speech_systems_combo.set_model(model)

        current = self._manager.get_current_server()
        found = False
        selected_server = None
        for i, row in enumerate(model):
            if row[0] == current:
                self._speech_systems_combo.set_active(i)
                selected_server = current
                found = True
                break

        if not found and len(model) > 0:
            self._speech_systems_combo.set_active(0)
            tree_iter = model.get_iter_first()
            if tree_iter:
                selected_server = model.get_value(tree_iter, 0)

        if selected_server:
            self._manager.set_current_server(selected_server)

        self._initializing = False
        self._populate_speech_synthesizers()

    def _populate_speech_synthesizers(self) -> None:
        """Populate the speech synthesizers combo."""

        self._initializing = True

        model = self._speech_synthesizers_combo.get_model()
        if not model:
            model = Gtk.ListStore(str)
        self._speech_synthesizers_combo.set_model(None)
        model.clear()

        available = self._manager.get_available_synthesizers()
        for synth_name in available:
            model.append([synth_name])

        self._speech_synthesizers_combo.set_model(model)

        current = self._manager.get_current_synthesizer()
        found = False
        selected_synth = None
        for i, row in enumerate(model):
            if row[0] == current:
                self._speech_synthesizers_combo.set_active(i)
                selected_synth = current
                found = True
                break

        if not found and len(model) > 0:
            self._speech_synthesizers_combo.set_active(0)
            tree_iter = model.get_iter_first()
            if tree_iter:
                selected_synth = model.get_value(tree_iter, 0)

        if selected_synth:
            self._manager.set_current_synthesizer(selected_synth)

        self._voice_families = self._manager.get_voice_families()
        self._initializing = False
        # Note: Voice widgets are created on-demand in dialogs, so we don't populate them here

    # pylint: disable-next=too-many-branches
    # pylint: disable-next=too-many-branches
    def _on_punctuation_changed(self, widget: Gtk.ComboBox) -> None:
        """Handle punctuation combo box change."""

        if self._initializing:
            return

        active = widget.get_active()
        if active < 0:
            return

        model = widget.get_model()
        tree_iter = model.get_iter(active)
        level = model.get_value(tree_iter, 1)

        gsettings_registry.get_registry().set_runtime_value(
            self._manager.SPEECH_SCHEMA,
            self._manager.KEY_PUNCTUATION_LEVEL,
            PunctuationStyle(level).string_name,
        )
        self._manager.update_punctuation_level()
        self._has_unsaved_changes = True

    def _on_capitalization_changed(self, widget: Gtk.ComboBox) -> None:
        """Handle capitalization combo box change."""

        if self._initializing:
            return

        active = widget.get_active()
        if active < 0:
            return

        model = widget.get_model()
        tree_iter = model.get_iter(active)
        style = model.get_value(tree_iter, 1)

        gsettings_registry.get_registry().set_runtime_value(
            self._manager.SPEECH_SCHEMA,
            self._manager.KEY_CAPITALIZATION_STYLE,
            CapitalizationStyle(style).string_name,
        )
        self._manager.update_capitalization_style()
        self._has_unsaved_changes = True

    def _on_speak_numbers_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle speak numbers as digits switch change."""
        if self._initializing:
            return
        self._manager.set_speak_numbers_as_digits(switch.get_active())
        self._has_unsaved_changes = True

    def _on_use_color_names_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle use color names switch change."""
        if self._initializing:
            return
        self._manager.set_use_color_names(switch.get_active())
        self._has_unsaved_changes = True

    def _on_enable_pause_breaks_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle enable pause breaks switch change."""
        if self._initializing:
            return
        self._manager.set_insert_pauses_between_utterances(switch.get_active())
        self._has_unsaved_changes = True

    def _on_use_pronunciation_dict_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle use pronunciation dictionary switch change."""
        if self._initializing:
            return
        self._manager.set_use_pronunciation_dictionary(switch.get_active())
        self._has_unsaved_changes = True

    def _on_auto_language_switching_content_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle auto language switching for document content switch change."""
        if self._initializing:
            return
        self._manager.set_auto_language_switching(switch.get_active())
        self._has_unsaved_changes = True

    def _on_auto_language_switching_ui_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle auto language switching for UI elements switch change."""
        if self._initializing:
            return
        self._manager.set_auto_language_switching_ui(switch.get_active())
        self._has_unsaved_changes = True

    def _on_only_switch_configured_languages_toggled(self, switch: Gtk.Switch, _state: Any) -> None:
        """Handle only-switch-configured-languages switch change."""
        if self._initializing:
            return
        self._manager.set_only_switch_configured_languages(switch.get_active())
        self._has_unsaved_changes = True

    def _on_speech_system_changed(self, widget: Gtk.ComboBox) -> None:
        """Handle speech system combo change."""

        if self._initializing:
            return

        active = widget.get_active()
        if active < 0:
            return

        model = widget.get_model()
        tree_iter = model.get_iter(active)
        server_name = model.get_value(tree_iter, 0)

        gsettings_registry.get_registry().set_runtime_value(
            self._manager.SPEECH_SCHEMA,
            self._manager.KEY_SPEECH_SERVER,
            server_name,
        )
        self._manager.set_current_server(server_name)

        self._populate_speech_synthesizers()
        self._has_unsaved_changes = True

    def _on_speech_synthesizer_changed(self, widget: Gtk.ComboBox) -> None:
        """Handle speech synthesizer combo change."""

        if self._initializing:
            return

        active = widget.get_active()
        if active < 0:
            return

        model = widget.get_model()
        tree_iter = model.get_iter(active)
        synth_name = model.get_value(tree_iter, 0)

        # Without the override, update_synthesizer() on script activation reverts
        # the live module to the dconf value before the user has clicked save.
        gsettings_registry.get_registry().set_runtime_value(
            self._manager.SPEECH_SCHEMA,
            self._manager.KEY_SYNTHESIZER,
            synth_name,
        )
        self._manager.set_current_synthesizer(synth_name)

        self._voice_families = self._manager.get_voice_families()
        self._families_sorted = False

        # When synthesizer changes, replace the old family with a default
        # from the new synthesizer. Without this, import_voice only writes
        # keys present in the ACSS dict, so old family values persist in dconf.
        default_family = self._voice_families[0] if self._voice_families else None
        for voice_type in speechserver.VoiceType:
            voice_acss = self._get_acss_for_voice_type(voice_type)
            if ACSS.FAMILY in voice_acss:
                del voice_acss[ACSS.FAMILY]
            if default_family is not None:
                voice_acss[ACSS.FAMILY] = default_family

        self._has_unsaved_changes = True

    def get_voice_for_type(self, voice_type: str) -> ACSS:
        """Returns a copy of the primary voice currently configured for voice_type."""

        return ACSS(self._voices.get(voice_type, self._voices[speechserver.VoiceType.DEFAULT]))

    def set_voice_for_type(self, voice_type: str, voice: ACSS) -> None:
        """Updates the primary voice for voice_type and marks the grid as changed."""

        configured = ACSS(voice)
        configured["established"] = True
        self._voices[voice_type] = configured
        self._has_unsaved_changes = True

    def get_all_voices(self) -> dict[str, ACSS]:
        """Returns copies of the staged primary voices keyed by voice type."""

        return {voice_type: ACSS(voice) for voice_type, voice in self._voices.items()}

    def get_voice_families(self) -> list[speechserver.VoiceFamily]:
        """Returns the full list of voice families from the current synthesizer."""

        return list(self._voice_families)

    def get_primary_voice_family(self) -> dict[str, str]:
        """Returns the family dict of the primary default voice."""

        return self._voices.get(speechserver.VoiceType.DEFAULT, ACSS({})).get(ACSS.FAMILY, {})

    def get_primary_language_codes(self) -> list[str]:
        """Returns the language codes of the primary default voice."""

        family = self._voices.get(speechserver.VoiceType.DEFAULT, ACSS({})).get(ACSS.FAMILY, {})
        lang = family.get(speechserver.VoiceFamily.LANG, "")
        if not lang:
            return []
        dialect = family.get(speechserver.VoiceFamily.DIALECT, "")
        full = f"{lang}-{dialect}" if dialect else ""
        codes = [lang]
        if lang.lower() != lang:
            codes.append(lang.lower())
        if full:
            codes.append(full)
            if full.lower() != full:
                codes.append(full.lower())
        return codes

    def get_available_languages(self) -> list[tuple[str, str]]:
        """Returns (lang_code, display_name) pairs from available voice families."""

        done: set[tuple[str, str]] = set()
        languages: list[tuple[str, str]] = []
        for family in self._voice_families:
            lang = family.get(speechserver.VoiceFamily.LANG, "")
            dialect = family.get(speechserver.VoiceFamily.DIALECT, "")
            if not lang or (lang, dialect) in done:
                continue
            if dialect and not language_utilities.is_standard_locale(lang, dialect):
                continue
            done.add((lang, dialect))
            code = f"{lang}-{dialect}" if dialect else lang
            display = language_utilities.get_language_display_name(lang, dialect)
            languages.append((code, display))
        return sorted(languages, key=lambda x: x[1])


# pylint: disable=no-member
# pylint: disable-next=too-many-instance-attributes
class VoiceTypesPreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """GtkGrid containing voice set selector and voice type buttons."""

    _PRIMARY_LABEL = guilabels.VOICE_SET_GLOBAL

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for voice type preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.LANGUAGE_VOICE_SETTINGS,
            panel_id="manual.voice-sets",
            summary="Use these settings to configure voice sets and the voice used for each "
            "kind of spoken content.",
            description=guilabels.VOICE_SET_INFO,
            schema="voice",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.LANGUAGE_VOICE_SETTINGS,
                    kind="group",
                    summary="Lists the global voice set and any language-specific voice sets.",
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_SET,
                            kind="choice",
                            summary="Selects the voice set to review or change.",
                            dynamic_values=True,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_SET_CREATE_NEW,
                            kind="button",
                            summary="Creates a voice set for a specific language.",
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label="Delete Voice Set",
                            kind="button",
                            summary="Removes the selected language-specific voice set.",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VOICE_VOICE_TYPE_SETTINGS,
                    kind="group",
                    summary="Lists the voice types in the selected voice set. Activate the "
                    "settings button for a voice type to change how that voice sounds.",
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_LANGUAGE,
                            kind="choice",
                            summary="Selects the language for the selected voice type.",
                            dynamic_values=True,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_PERSON,
                            kind="choice",
                            summary="Selects the synthesizer voice for the selected voice type.",
                            schema="voice",
                            key="family-name",
                            dynamic_values=True,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_RATE,
                            kind="integer",
                            schema="voice",
                            key="rate",
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_PITCH,
                            kind="number",
                            schema="voice",
                            key="pitch",
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_INFLECTION,
                            kind="number",
                            schema="voice",
                            key="pitch-range",
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.VOICE_VOLUME,
                            kind="number",
                            schema="voice",
                            key="volume",
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, voices_grid: VoicesPreferencesGrid) -> None:
        super().__init__(guilabels.VOICE_TYPES)
        self._voices_grid = voices_grid
        self._manager = voices_grid._manager
        self._staged_configs: dict[str, dict[str, ACSS | None]] = {}
        self._deleted_sets: set[str] = set()
        self._voice_set_combo: Gtk.ComboBoxText
        self._delete_button: Gtk.Button
        self._buttons_listbox: preferences_grid_base.PreferencesFocusManagedListBox
        self._build()

    def _get_voice_set_items(self) -> list[tuple[str, str]]:
        """Returns (id, label) pairs for the voice set combo."""

        primary_family = self._voices_grid.get_primary_voice_family()
        primary_lang_display = ""
        if primary_family:
            lang = primary_family.get(speechserver.VoiceFamily.LANG, "")
            dialect = primary_family.get(speechserver.VoiceFamily.DIALECT, "")
            if lang:
                primary_lang_display = language_utilities.get_language_display_name(lang, dialect)
        if primary_lang_display:
            primary_label = f"{self._PRIMARY_LABEL}: {primary_lang_display}"
        else:
            primary_label = self._PRIMARY_LABEL
        items = [(gsettings_registry.PRIMARY_VOICE_SET, primary_label)]
        names = set(self._manager.get_voice_set_names())
        names.update(self._staged_configs.keys())
        names -= self._deleted_sets
        items.extend(
            (name, language_utilities.get_language_display_name(name)) for name in sorted(names)
        )
        return items

    def _build(self) -> None:
        row = 0

        msg = f"{guilabels.VOICE_SET_INFO} {guilabels.VOICE_SET_INFO_COMMANDS}"
        info_listbox = orca_gui_helpers.create_info_listbox(msg)
        self.attach(info_listbox, 0, row, 1, 1)
        row += 1

        header_box, self._add_button = orca_gui_helpers.create_heading_action_box(
            guilabels.LANGUAGE_VOICE_SETTINGS,
            "list-add-symbolic",
            guilabels.VOICE_SET_CREATE_NEW,
            self._on_add_voice_set,
        )

        self.attach(header_box, 0, row, 1, 1)
        row += 1

        items = self._get_voice_set_items()
        self._voice_set_combo = Gtk.ComboBoxText()
        for item_id, display in items:
            self._voice_set_combo.append(item_id, display)
        if items:
            self._voice_set_combo.set_active(0)

        self._delete_button = Gtk.Button.new_from_icon_name(
            "edit-delete-symbolic", Gtk.IconSize.BUTTON
        )
        self._delete_button.set_valign(Gtk.Align.CENTER)
        self._delete_button.set_sensitive(False)
        self._delete_button.connect("clicked", self._on_delete_voice_set)

        combo_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        combo_box.pack_start(self._voice_set_combo, False, False, 0)
        combo_box.pack_start(self._delete_button, False, False, 0)

        selector_row, _hbox, _label = orca_gui_helpers.create_row_structure(
            False, guilabels.VOICE_SET, combo_box, label_halign=Gtk.Align.START
        )

        selector_listbox = preferences_grid_base.PreferencesFocusManagedListBox()
        selector_listbox.add_row_with_widget(selector_row, self._voice_set_combo)
        self.attach(selector_listbox, 0, row, 1, 1)
        row += 1

        self._voice_types_section, buttons_content, buttons_heading = (
            orca_gui_helpers.create_section_box(
                guilabels.VOICE_VOICE_TYPE_SETTINGS,
                margin_top=12,
            )
        )
        self._buttons_listbox = preferences_grid_base.PreferencesFocusManagedListBox()
        orca_gui_helpers.add_labelled_by(self._buttons_listbox, buttons_heading)
        orca_gui_helpers.add_section_content(
            self._voice_types_section,
            buttons_content,
            self._buttons_listbox,
        )
        self.attach(self._voice_types_section, 0, row, 1, 1)

        self._voice_set_combo.connect("changed", self._on_voice_set_changed)
        self._rebuild_buttons()
        self.show_all()  # pylint: disable=no-member

    def _rebuild_buttons(self) -> None:
        """Rebuild voice type buttons for the selected voice set."""

        for child in self._buttons_listbox.get_children():
            self._buttons_listbox.remove(child)

        voice_set = self._voice_set_combo.get_active_id()
        is_primary = voice_set == gsettings_registry.PRIMARY_VOICE_SET
        self._delete_button.set_sensitive(not is_primary)

        for vt in speechserver.VoiceType:
            label = guilabels.VOICE_TYPE_LABELS.get(vt, vt)

            def _make_handler(voice_type: str = vt):
                return lambda _btn: self._show_voice_dialog(voice_type, voice_set)

            row, button, _label = orca_gui_helpers.create_button_row(
                label,
                "applications-system-symbolic",
                _make_handler(),
                include_top_separator=False,
            )
            accessible_name = guilabels.VOICE_TYPE_SETTINGS % label
            button.set_tooltip_text(accessible_name)
            accessible = button.get_accessible()
            if accessible:
                accessible.set_name(accessible_name)
            self._buttons_listbox.add_row_with_widget(row, button)

        self._buttons_listbox.show_all()  # pylint: disable=no-member

    def _on_voice_set_changed(self, _combo: Gtk.ComboBoxText) -> None:
        """Handle voice set combo selection change."""
        self._rebuild_buttons()

    def _get_addable_languages(self) -> list[tuple[str, str]]:
        """Returns languages not yet configured as voice sets."""

        existing = set(self._manager.get_voice_set_names()) - self._deleted_sets
        existing.update(self._staged_configs.keys())

        existing.update(self._voices_grid.get_primary_language_codes())

        languages: dict[str, str] = {}
        seen_display_names: set[str] = set()
        for lang, dialect in language_utilities.get_known_language_codes():
            code = f"{lang}-{dialect}" if dialect else lang
            if code in existing or code in languages:
                continue
            display = language_utilities.get_language_display_name(lang, dialect)
            if display == code:
                continue
            if dialect and not language_utilities.is_standard_locale(lang, dialect):
                continue
            if display in seen_display_names:
                continue
            seen_display_names.add(display)
            languages[code] = display

        return sorted(languages.items(), key=lambda x: x[1])

    def _families_for_language(self, lang_code: str) -> list[speechserver.VoiceFamily]:
        """Returns the voice families matching lang_code, in synthesizer order."""

        match_lang, _sep, match_dialect = lang_code.lower().partition("-")
        families = []
        for family in self._voices_grid.get_voice_families():
            lang = family.get(speechserver.VoiceFamily.LANG, "").lower()
            dialect = family.get(speechserver.VoiceFamily.DIALECT, "").lower()
            if lang == match_lang and (not match_dialect or dialect == match_dialect):
                families.append(family)
        return families

    def _default_voice_set_config(self, lang_code: str) -> ACSS:
        """Returns a complete default voice config for a newly created voice set."""

        props: dict[str, Any] = {
            ACSS.RATE: ACSS.settings[ACSS.RATE],
            ACSS.AVERAGE_PITCH: ACSS.settings[ACSS.AVERAGE_PITCH],
            ACSS.PITCH_RANGE: ACSS.settings[ACSS.PITCH_RANGE],
            ACSS.GAIN: ACSS.settings[ACSS.GAIN],
        }
        if families := self._families_for_language(lang_code):
            props[ACSS.FAMILY] = dict(families[0])
        return ACSS(props)

    def _on_add_voice_set(self, _button: Gtk.Button) -> None:
        """Handle add voice set button click."""

        available = self._get_addable_languages()
        if not available:
            return

        dialog, ok_button = orca_gui_helpers.create_header_bar_dialog(
            guilabels.LANGUAGE_VOICE_SETTINGS,
            guilabels.BTN_CANCEL,
            guilabels.BTN_OK,
            transient_for=self.get_toplevel(),
            width=400,
        )
        content_area = dialog.get_content_area()
        lang_row, lang_combo, _label = orca_gui_helpers.create_combo_box_text_row(
            guilabels.VOICE_LANGUAGE,
            available,
            include_top_separator=False,
        )

        listbox = orca_gui_helpers.FocusManagedListBox()
        listbox.add_row_with_widget(lang_row, lang_combo)
        content_area.pack_start(listbox, False, False, 0)

        def on_response(dlg: Gtk.Dialog, response_id: int) -> None:
            if response_id not in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
                lang_code = lang_combo.get_active_id()
                if lang_code:
                    self._staged_configs.setdefault(lang_code, {})[
                        speechserver.VoiceType.DEFAULT
                    ] = self._default_voice_set_config(lang_code)
                    self._deleted_sets.discard(lang_code)
                    self._refresh_voice_set_combo()
                    self._select_voice_set(lang_code)
                    self._has_unsaved_changes = True
            dlg.destroy()

        dialog.connect("response", on_response)
        dialog.show_all()  # pylint: disable=no-member
        ok_button.grab_default()

    def _on_delete_voice_set(self, _button: Gtk.Button) -> None:
        """Handle delete voice set button click."""

        voice_set = self._voice_set_combo.get_active_id()
        if not voice_set or voice_set == gsettings_registry.PRIMARY_VOICE_SET:
            return

        parent = self.get_toplevel()  # pylint: disable=no-member
        dialog = Gtk.MessageDialog(
            transient_for=parent if parent.is_toplevel() else None,
            modal=True,
            message_type=Gtk.MessageType.WARNING,
            buttons=Gtk.ButtonsType.YES_NO,
            text=guilabels.VOICE_SET_DELETE_CONFIRMATION
            % language_utilities.get_language_display_name(voice_set),
        )
        response = dialog.run()
        dialog.destroy()

        if response != Gtk.ResponseType.YES:
            return

        self._deleted_sets.add(voice_set)
        self._staged_configs.pop(voice_set, None)
        self._has_unsaved_changes = True

        self._refresh_voice_set_combo()
        self._voice_set_combo.set_active(0)

    def _refresh_voice_set_combo(self) -> None:
        """Refresh the voice set combo with current voice sets."""

        self._voice_set_combo.remove_all()
        items = self._get_voice_set_items()
        for item_id, display in items:
            self._voice_set_combo.append(item_id, display)
        if items:
            self._voice_set_combo.set_active(0)

    def _select_voice_set(self, voice_set: str) -> None:
        """Select a voice set in the combo by id."""

        model = self._voice_set_combo.get_model()
        for i, row in enumerate(model):
            if row[1] == voice_set:
                self._voice_set_combo.set_active(i)
                return

    def _show_voice_dialog(self, voice_type: str, voice_set: str) -> None:
        """Show a dialog for editing a voice type within a voice set, with live preview."""

        label = guilabels.VOICE_TYPE_LABELS.get(voice_type, voice_type)
        is_primary = voice_set == gsettings_registry.PRIMARY_VOICE_SET
        if is_primary:
            title = label
            config = self._voices_grid.get_voice_for_type(voice_type)
        else:
            voice_set_display = language_utilities.get_language_display_name(voice_set)
            title = f"{label} ({voice_set_display})"
            staged = self._staged_configs.get(voice_set, {}).get(voice_type)
            config = staged or self._manager.get_voice_properties(voice_type, voice_set=voice_set)

        # Edits are applied to the live voice so they can be heard while the dialog is open;
        # the configured voice is restored when it closes.

        dialog, ok_button = orca_gui_helpers.create_header_bar_dialog(
            title,
            guilabels.BTN_CANCEL,
            guilabels.BTN_OK,
            transient_for=self.get_toplevel(),
            width=500,
        )

        content_area = dialog.get_content_area()
        voice_listbox = orca_gui_helpers.FocusManagedListBox()
        combo_size_group = orca_gui_helpers.create_horizontal_size_group()

        available_languages = self._voices_grid.get_available_languages()
        person_choices: list[dict[str, Any]] = []

        voice_lang_row, voice_lang_combo, _vl_label = orca_gui_helpers.create_combo_box_text_row(
            guilabels.VOICE_LANGUAGE,
            available_languages,
            include_top_separator=False,
        )
        combo_size_group.add_widget(voice_lang_combo)
        voice_listbox.add_row_with_widget(voice_lang_row, voice_lang_combo)

        person_row, person_combo, _pl = orca_gui_helpers.create_combo_box_text_row(
            guilabels.VOICE_PERSON, [], include_top_separator=False
        )
        combo_size_group.add_widget(person_combo)
        voice_listbox.add_row_with_widget(person_row, person_combo)

        def populate_persons(voice_language: str) -> None:
            person_choices.clear()
            person_combo.remove_all()
            for family in self._families_for_language(voice_language):
                name = family.get(speechserver.VoiceFamily.NAME, "")
                variant = family.get(speechserver.VoiceFamily.VARIANT, "")
                display = variant if variant and variant not in ("none", "None") else name
                person_combo.append_text(display)
                person_choices.append(family)
            if person_choices:
                person_combo.set_active(0)

        def on_voice_lang_changed(_combo: Gtk.ComboBoxText) -> None:
            lang_id = voice_lang_combo.get_active_id() or ""
            populate_persons(lang_id)

        voice_lang_combo.connect("changed", on_voice_lang_changed)

        # Select the language that matches the config, or the voice set's language
        config_family = config.get(ACSS.FAMILY, {})
        config_lang = config_family.get(speechserver.VoiceFamily.LANG, "")
        config_dialect = config_family.get(speechserver.VoiceFamily.DIALECT, "")
        if config_lang:
            target = (
                f"{config_lang}-{config_dialect}".lower() if config_dialect else config_lang.lower()
            )
        else:
            target = voice_set.lower()

        partial_match = -1
        for i, (lc, _ld) in enumerate(available_languages):
            if lc.lower() == target:
                voice_lang_combo.set_active(i)
                break
            if partial_match < 0 and lc.lower().split("-")[0] == target.split("-")[0]:
                partial_match = i
        else:
            if partial_match >= 0:
                voice_lang_combo.set_active(partial_match)

        populate_persons(voice_lang_combo.get_active_id() or "")
        config_name = config_family.get(speechserver.VoiceFamily.NAME, "")
        for i, fam in enumerate(person_choices):
            if fam.get(speechserver.VoiceFamily.NAME, "") == config_name:
                person_combo.set_active(i)
                break

        current_rate = config.get(ACSS.RATE, 50)
        rate_row, rate_scale, _rl = orca_gui_helpers.create_range_slider_row(
            guilabels.VOICE_RATE,
            current_rate,
            0,
            100,
            include_top_separator=False,
        )
        voice_listbox.add_row_with_widget(rate_row, rate_scale)

        current_pitch = config.get(ACSS.AVERAGE_PITCH, 5.0)
        pitch_row, pitch_scale, _pl = orca_gui_helpers.create_range_slider_row(
            guilabels.VOICE_PITCH,
            current_pitch,
            0,
            10,
            include_top_separator=False,
            digits=1,
        )
        voice_listbox.add_row_with_widget(pitch_row, pitch_scale)

        current_pr = config.get(ACSS.PITCH_RANGE, 5.0)
        pr_row, pr_scale, _prl = orca_gui_helpers.create_range_slider_row(
            guilabels.VOICE_INFLECTION,
            current_pr,
            0,
            10,
            include_top_separator=False,
            digits=1,
        )
        voice_listbox.add_row_with_widget(pr_row, pr_scale)

        current_vol = config.get(ACSS.GAIN, 10.0)
        vol_row, vol_scale, _vl = orca_gui_helpers.create_range_slider_row(
            guilabels.VOICE_VOLUME,
            current_vol,
            0,
            10,
            include_top_separator=False,
            digits=1,
        )
        voice_listbox.add_row_with_widget(vol_row, vol_scale)

        def update_live_voice(*_args: Any) -> None:
            active = person_combo.get_active()
            if not 0 <= active < len(person_choices):
                return
            self._manager.apply_live_voice(
                ACSS(
                    {
                        ACSS.FAMILY: dict(person_choices[active]),
                        ACSS.RATE: int(rate_scale.get_value()),
                        ACSS.AVERAGE_PITCH: pitch_scale.get_value(),
                        ACSS.PITCH_RANGE: pr_scale.get_value(),
                        ACSS.GAIN: vol_scale.get_value(),
                    }
                )
            )

        person_combo.connect("changed", update_live_voice)
        for scale in (rate_scale, pitch_scale, pr_scale, vol_scale):
            scale.connect("value-changed", update_live_voice)

        # Apply the configured voice immediately so it is heard on entry, not only on edit.
        # This matters for sets whose voice differs from the global one.
        update_live_voice()

        def on_response(dlg: Gtk.Dialog, response_id: int) -> None:
            confirmed = response_id not in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT)
            if confirmed:
                props: dict[str, Any] = {
                    ACSS.RATE: int(rate_scale.get_value()),
                    ACSS.AVERAGE_PITCH: pitch_scale.get_value(),
                    ACSS.PITCH_RANGE: pr_scale.get_value(),
                    ACSS.GAIN: vol_scale.get_value(),
                }
                active = person_combo.get_active()
                if 0 <= active < len(person_choices):
                    props[ACSS.FAMILY] = dict(person_choices[active])
                if is_primary:
                    self._voices_grid.set_voice_for_type(voice_type, ACSS(props))
                else:
                    self._staged_configs.setdefault(voice_set, {})[voice_type] = ACSS(props)
                    self._has_unsaved_changes = True
            # Make the staged global voices live (each by type), replacing the audition
            # override. So an OK'd default/system/etc. voice stays in effect; a set's voice
            # was staged separately and only applies when that set is active.
            self._manager.restore_live_voices(self._voices_grid.get_all_voices())
            dlg.destroy()

        def on_parent_destroy(*_args: Any) -> None:
            if dialog.get_property("visible"):
                dialog.response(Gtk.ResponseType.DELETE_EVENT)

        dialog.connect("response", on_response)
        self.get_toplevel().connect("destroy", on_parent_destroy)  # pylint: disable=no-member
        content_area.pack_start(voice_listbox, True, True, 0)
        dialog.show_all()  # pylint: disable=no-member
        ok_button.grab_default()

    def save_settings(self) -> dict[str, dict | list | int | str | bool]:
        """Writes staged voice set changes to dconf."""

        registry = gsettings_registry.get_registry()
        profile = registry.get_active_profile()

        for voice_set in self._deleted_sets:
            for vt in speechserver.VoiceType:
                sub = gsettings_registry.get_registry().voice_set_sub_path(vt, voice_set)
                gs = registry.get_settings("voice", profile, sub)
                if gs is not None:
                    for key in gs.list_keys():
                        if gs.get_user_value(key) is not None:
                            gs.reset(key)

        for voice_set, voice_types in self._staged_configs.items():
            for voice_type, config in voice_types.items():
                if config is not None:
                    self._manager.set_voice_set_properties(voice_type, voice_set, config)

        self._staged_configs.clear()
        self._deleted_sets.clear()
        self._has_unsaved_changes = False
        self._manager.refresh_voice_set_commands()
        return {}

    def has_changes(self) -> bool:
        """Return True if there are unsaved voice set changes."""

        return self._has_unsaved_changes

    def reload(self) -> None:
        """Reload voice set state and refresh UI."""

        self._has_unsaved_changes = False
        self.refresh()

    def refresh(self) -> None:
        """Refresh voice set combo to reflect current state."""

        self._staged_configs.clear()
        self._deleted_sets.clear()
        if self._voice_set_combo is not None:
            self._refresh_voice_set_combo()
            self._rebuild_buttons()


# pylint: enable=no-member



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2023 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for Where Am I."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .where_am_i_presenter import WhereAmIPresenter


def get_commands(owner: WhereAmIPresenter) -> list[Command]:
    """Returns commands for Where Am I."""

    kb_orca_f = keybindings.KeyBinding("f", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_f_2 = keybindings.KeyBinding(
        "f",
        keybindings.ORCA_MODIFIER_MASK,
        click_count=2,
    )
    kb_orca_e = keybindings.KeyBinding("e", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_shift_up = keybindings.KeyBinding("Up", keybindings.ORCA_SHIFT_MODIFIER_MASK)

    kb_orca_equal = keybindings.KeyBinding("equal", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_kp_enter = keybindings.KeyBinding("KP_Enter", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_kp_enter_2 = keybindings.KeyBinding(
        "KP_Enter",
        keybindings.ORCA_MODIFIER_MASK,
        click_count=2,
    )
    kb_kp_enter = keybindings.KeyBinding("KP_Enter", keybindings.NO_MODIFIER_MASK)
    kb_kp_enter_2 = keybindings.KeyBinding(
        "KP_Enter",
        keybindings.NO_MODIFIER_MASK,
        click_count=2,
    )

    kb_orca_slash = keybindings.KeyBinding("slash", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_slash_2 = keybindings.KeyBinding(
        "slash",
        keybindings.ORCA_MODIFIER_MASK,
        click_count=2,
    )
    kb_orca_return = keybindings.KeyBinding("Return", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_return_2 = keybindings.KeyBinding(
        "Return",
        keybindings.ORCA_MODIFIER_MASK,
        click_count=2,
    )

    return [
        KeyboardCommand(
            "readCharAttributesHandler",
            owner.present_character_attributes,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_CHARACTER_FORMATTING,
            desktop_keybinding=kb_orca_f,
            laptop_keybinding=kb_orca_f,
        ),
        KeyboardCommand(
            "show_character_attributes",
            owner.show_character_attributes,
            owner.GROUP_LABEL,
            cmdnames.SHOW_CHARACTER_FORMATTING,
            desktop_keybinding=kb_orca_f_2,
            laptop_keybinding=kb_orca_f_2,
        ),
        KeyboardCommand(
            "presentSizeAndPositionHandler",
            owner.present_size_and_position,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_SIZE_AND_POSITION,
        ),
        KeyboardCommand(
            "getTitleHandler",
            owner.present_title,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_TITLE,
            desktop_keybinding=kb_orca_kp_enter,
            laptop_keybinding=kb_orca_slash,
        ),
        KeyboardCommand(
            "getStatusBarHandler",
            owner.present_status_bar,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_STATUS_BAR,
            desktop_keybinding=kb_orca_kp_enter_2,
            laptop_keybinding=kb_orca_slash_2,
        ),
        KeyboardCommand(
            "present_default_button",
            owner.present_default_button,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_DEFAULT_BUTTON,
            desktop_keybinding=kb_orca_e,
            laptop_keybinding=kb_orca_e,
        ),
        KeyboardCommand(
            "present_cell_formula",
            owner.present_cell_formula,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_CELL_FORMULA,
            desktop_keybinding=kb_orca_equal,
        ),
        KeyboardCommand(
            "whereAmIBasicHandler",
            owner.where_am_i_basic,
            owner.GROUP_LABEL,
            cmdnames.WHERE_AM_I_BASIC,
            desktop_keybinding=kb_kp_enter,
            laptop_keybinding=kb_orca_return,
        ),
        KeyboardCommand(
            "whereAmIDetailedHandler",
            owner.where_am_i_detailed,
            owner.GROUP_LABEL,
            cmdnames.WHERE_AM_I_DETAILED,
            desktop_keybinding=kb_kp_enter_2,
            laptop_keybinding=kb_orca_return_2,
        ),
        KeyboardCommand(
            "whereAmILinkHandler",
            owner.present_link,
            owner.GROUP_LABEL,
            cmdnames.WHERE_AM_I_LINK,
        ),
        KeyboardCommand(
            "whereAmISelectionHandler",
            owner.present_selection,
            owner.GROUP_LABEL,
            cmdnames.WHERE_AM_I_SELECTION,
            desktop_keybinding=kb_orca_shift_up,
            laptop_keybinding=kb_orca_shift_up,
        ),
    ]
